from _typeshed import Incomplete
from mne import BaseEpochs
from mne.io import BaseRaw
from mne.preprocessing import ICA
from qtpy.QtWidgets import QMainWindow

from mne_icalabel.config import ICA_LABELS_TO_MNE as ICA_LABELS_TO_MNE

class ICAComponentLabeler(QMainWindow):
    """Qt GUI to annotate components.

    Parameters
    ----------
    inst : Raw | Epochs
    ica : ICA
    show : bool
    """

    _inst: Incomplete
    _ica: Incomplete
    _labels: Incomplete
    selected_labels: Incomplete
    _selected_component: int

    def __init__(
        self, inst: BaseRaw | BaseEpochs, ica: ICA, show: bool = True
    ) -> None: ...
    def _save_labels(self) -> None:
        """Save the selected labels to the ICA instance."""
    _central_widget: Incomplete
    _components_listWidget: Incomplete
    _labels_buttonGroup: Incomplete
    _mpl_figures: Incomplete
    _mpl_widgets: Incomplete
    _timeSeries_widget: Incomplete

    def _load_ui(self) -> None:
        """Prepare the GUI.

        Widgets
        -------
        self._components_listWidget
        self._labels_buttonGroup
        self._mpl_widgets (dict)
            - topomap
            - psd
        self._timeSeries_widget

        Matplotlib figures
        ------------------
        self._mpl_figures (dict)
            - topomap
            - psd
        """

    @staticmethod
    def _check_inst_ica(inst: BaseRaw | BaseEpochs, ica: ICA) -> None:
        """Check if the ICA was fitted."""

    @property
    def inst(self) -> BaseRaw | BaseEpochs:
        """Instance on which the ICA has been fitted."""

    @property
    def ica(self) -> ICA:
        """Fitted ICA decomposition."""

    @property
    def n_components_(self) -> int:
        """The number of fitted components."""

    @property
    def labels(self) -> list[str]:
        """List of valid labels."""

    @property
    def selected_component(self) -> int:
        """IC selected and displayed."""

    def _connect_signals_to_slots(self) -> None:
        """Connect all the signals and slots of the GUI."""

    def _components_listWidget_clicked(self) -> None:
        """Update the plots and the saved labels accordingly."""

    def _update_selected_labels(self) -> None:
        """Update the labels saved."""

    def _reset(self) -> None:
        """Remove the label from the current component."""

    def _reset_buttons(self) -> None:
        """Reset all buttons."""

    def closeEvent(self, event) -> None:
        """Clean up upon closing the window.

        Update the labels since the user might have selected one for the
        currently being displayed IC.
        """



import re
from pathlib import Path

import pytest
from mne.utils import verbose

from tmseegpy.mne_icalabel.utils._docs import _KEYS_MNE, copy_doc, docdict, fill_doc


def test_fill_doc_function():
    """Test decorator to fill docstring on functions."""

    # test filling docstring
    @fill_doc
    def foo(verbose):
        """My doc.

        Parameters
        ----------
        %(verbose)s
        """

    assert "verbose : bool | str | int | None" in foo.__doc__

    # test filling empty-docstring
    @fill_doc
    def foo():
        pass

    assert foo.__doc__ is None

    # test filling docstring with invalid key
    with pytest.raises(RuntimeError, match="Error documenting"):

        @fill_doc
        def foo(verbose):
            """My doc.

            Parameters
            ----------
            %(invalid_key)s
            """

    # test filling docstring with verbose
    @fill_doc
    @verbose
    def foo(verbose=None):
        """My doc.

        Parameters
        ----------
        %(verbose)s
        """

    assert "verbose : bool | str | int | None" in foo.__doc__


def test_fill_doc_class():
    """Test decorator to fill docstring on classes."""

    @fill_doc
    class Foo:
        """My doc.

        Parameters
        ----------
        %(verbose)s
        """

        def __init__(self, verbose=None):
            pass

        @fill_doc
        def method(self, verbose=None):
            """My method doc.

            Parameters
            ----------
            %(verbose)s
            """

        @fill_doc
        @verbose
        def method_decorated(self, verbose=None):
            """My method doc.

            Parameters
            ----------
            %(verbose)s
            """

        @staticmethod
        @fill_doc
        @verbose
        def method_decorated_static(verbose=None):
            """My method doc.

            Parameters
            ----------
            %(verbose)s
            """

    assert "verbose : bool | str | int | None" in Foo.__doc__
    assert "verbose : bool | str | int | None" in Foo.method.__doc__
    assert "verbose : bool | str | int | None" in Foo.method_decorated.__doc__
    assert "verbose : bool" in Foo.method_decorated_static.__doc__
    foo = Foo()
    assert "verbose : bool | str | int | None" in foo.__doc__
    assert "verbose : bool | str | int | None" in foo.method.__doc__
    assert "verbose : bool | str | int | None" in foo.method_decorated.__doc__


def test_copy_doc_function():
    """Test decorator to copy docstring on functions."""

    # test copy of docstring
    def foo(x, y):
        """My doc."""

    @copy_doc(foo)
    def foo2(x, y):
        pass

    @copy_doc(foo)
    def foo3(x, y):
        """Doc of foo3."""

    assert foo.__doc__ == foo2.__doc__
    assert foo.__doc__ + "Doc of foo3." == foo3.__doc__

    # test copy of docstring from a function without docstring
    def foo(x, y):
        pass

    with pytest.raises(RuntimeError, match="The docstring from foo could not"):

        @copy_doc(foo)
        def foo2(x, y):
            pass

    # test copy docstring of decorated function
    @verbose
    def foo(verbose=None):
        """My doc."""

    @copy_doc(foo)
    def foo2(verbose=None):
        pass

    @copy_doc(foo)
    @verbose
    def foo3(verbose=None):
        pass

    assert foo.__doc__ == foo2.__doc__
    assert foo.__doc__ == foo3.__doc__


def test_copy_doc_class():
    """Test decorator to copy docstring on classes."""

    class Foo:
        """My doc."""

        def __init__(self):
            pass

        def method1(self):
            """Super 101 doc."""

    @copy_doc(Foo)
    class Foo2:
        def __init__(self):
            pass

        @copy_doc(Foo.method1)
        def method2(self):
            pass

        @copy_doc(Foo.method1)
        @verbose
        def method3(self, verbose=None):
            pass

        @staticmethod
        @copy_doc(Foo.method1)
        @verbose
        def method4(verbose=None):
            pass

    assert Foo.__doc__ == Foo2.__doc__
    assert Foo.method1.__doc__ == Foo2.method2.__doc__
    assert Foo.method1.__doc__ == Foo2.method3.__doc__
    assert Foo.method1.__doc__ == Foo2.method4.__doc__
    foo = Foo()
    foo2 = Foo2()
    assert foo.__doc__ == foo2.__doc__
    assert foo.method1.__doc__ == foo2.method2.__doc__
    assert foo.method1.__doc__ == foo2.method3.__doc__
    assert foo.method1.__doc__ == foo2.method4.__doc__


def test_docdict_order():
    """Test that docdict is alphabetical."""
    # read the file as text, and get entries via regex
    docdict_ = docdict.copy()
    for key in _KEYS_MNE:
        del docdict_[key]
    docs_path = Path(__file__).parents[1] / "_docs.py"
    assert docs_path.is_file()
    with open(docs_path, encoding="UTF-8") as fid:
        docs = fid.read()
    entries = re.findall(r'docdict\[(?:\n    )?["\'](.+)["\']\n?\] = ', docs)
    # test length, uniqueness and order
    assert len(docdict_) == len(entries)
    assert sorted(entries) == entries



ICLABEL_NUMERICAL_TO_STRING: dict = {
    0: "brain",
    1: "muscle artifact",
    2: "eye blink",
    3: "heart beat",
    4: "line noise",
    5: "channel noise",
    6: "other",
}

ICLABEL_STRING_TO_NUMERICAL: dict = {
    val: key for key, val in ICLABEL_NUMERICAL_TO_STRING.items()
}



from __future__ import annotations  # c.f. PEP 563, PEP 649

from typing import TYPE_CHECKING

from mne.utils import _check_pandas_installed

from ..config import ICA_LABELS_TO_MNE
from ..iclabel._config import ICLABEL_STRING_TO_NUMERICAL
from ..utils._imports import import_optional_dependency

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Union

    from mne.preprocessing import ICA


def write_components_tsv(ica: ICA, fname: Union[str, Path]):
    """Write channels tsv file for ICA components.

    Will create an accompanying JSON sidecar to explain the
    additional columns created in the channels tsv file for the
    ICA component labeling.

    Parameters
    ----------
    ica : ICA
        An instance of the fitted ICA.
    fname : str | Path
        The output filename.

    Notes
    -----
    Components are stored in a ``.tsv`` file essentially in the same manner as
    ``channels.tsv`` files for BIDS-EEG data. For more information, see the
    `BIDS specification <https://bids-specification.readthedocs.io/en/stable/04-modality-specific-files/03-electroencephalography.html#channels-description-_channelstsv>`_
    for EEG channels metadata.

    Storage of ICA annotations as a ``.tsv`` file is currently experimental in the
    context of BIDS-EEG Derivatives. The API and functionality is subject to change
    as the community converges on the specification of BIDS-Derivatives.
    """  # noqa: E501
    import_optional_dependency("mne_bids")

    from mne_bids import BIDSPath, get_bids_path_from_fname, update_sidecar_json
    from mne_bids.write import _write_json

    pd = _check_pandas_installed(strict=True)

    # ensure the filename is a Path object
    if not isinstance(fname, BIDSPath):
        fname = get_bids_path_from_fname(fname)

    # initialize status, description and IC type
    status = ["good"] * ica.n_components_
    status_description = ["n/a"] * ica.n_components_
    ic_type = ["n/a"] * ica.n_components_

    # extract the component labels if they are present in the ICA instance
    if ica.labels_:
        for label, comps in ica.labels_.items():
            this_status = "good" if label == "brain" else "bad"
            if label in ICA_LABELS_TO_MNE.values():
                for comp in comps:
                    status[comp] = this_status
                    ic_type[comp] = label

    # Create TSV.
    tsv_data = pd.DataFrame(
        dict(
            component=list(range(ica.n_components_)),
            type=["ica"] * ica.n_components_,
            description=["Independent Component"] * ica.n_components_,
            status=status,
            status_description=status_description,
            annotate_method=["n/a"] * ica.n_components_,
            annotate_author=["n/a"] * ica.n_components_,
            ic_type=ic_type,
        )
    )
    # make sure parent directories exist
    fname.mkdir(exist_ok=True)
    tsv_data.to_csv(fname, sep="\t", index=False, encoding="utf-8")

    # create an accompanying JSON file describing the corresponding
    # extra columns for ICA labeling
    component_json = {
        "annotate_method": "Method used for annotating components (e.g. manual, "
        + "iclabel)",
        "annotate_author": "The name of the person who ran the annotation",
        "ic_type": "The type of annotation must be one of ['brain', "
        "'muscle artifact', 'eye blink', 'heart beat', 'line noise', "
        "'channel noise', 'other']",
    }
    fname = fname.copy().update(extension=".json")

    if not fname.fpath.exists():
        _write_json(fname, component_json)
    else:
        update_sidecar_json(fname, component_json)


def mark_component(
    component: int,
    fname: Union[str, Path],
    method: str,
    label: str,
    author: str,
    strict_label: bool = True,
):
    """Mark a component with a label.

    Parameters
    ----------
    component : int
        The component to mark.
    fname : Union[str, Path]
        The filename for the BIDS filepath.
    method : str
        The method to use. Must be 'manual', or one of ['iclabel'].
    label : str
        The label of the ICA component. Must be one of ['brain',
        'muscle artifact', 'eye blink', 'heart beat', 'line noise',
        'channel noise', 'other'].
    author : str
        The annotating author.
    strict_label : bool
        Whether to raise an error if ``label`` is not an accepted value.
        Default is True.

    Notes
    -----
    Storage of ICA annotations as a ``.tsv`` file is currently experimental in the
    context of BIDS-EEG Derivatives. The API and functionality is subject to change
    as the community converges on the specification of BIDS-Derivatives.
    """
    from mne_bids import BIDSPath, get_bids_path_from_fname

    pd = _check_pandas_installed(strict=True)

    if not isinstance(fname, BIDSPath):
        fname = get_bids_path_from_fname(fname)

    # read the file
    with open(fname) as fin:
        tsv_data = pd.read_csv(
            fin,
            sep="\t",
            index_col=None,
            dtype={
                "component": int,
                "type": str,
                "description": str,
                "status": str,
                "status_description": str,
                "annotate_method": str,
                "annotate_author": str,
                "ic_type": str,
            },
        )

    if component not in tsv_data["component"]:
        raise ValueError(f"Component {component} not in tsv data of {fname}.")

    # check label is correct
    if label not in ICLABEL_STRING_TO_NUMERICAL.keys() and strict_label:
        raise ValueError(
            f"IC annotated label {label} is not a valid label. "
            f"Please use one of {list(ICLABEL_STRING_TO_NUMERICAL.keys())}."
        )
    if label == "brain":
        status = "good"
    else:
        status = "bad"

    # check that method is one of the allowed values
    description = ""
    if method == "manual":
        description += "Manually-annotated: "
    else:
        description += f"Auto-detected with {method}: "
    description += f"{label}"

    # Read sidecar and create required columns if they do not exist.
    if "status" not in tsv_data:
        tsv_data["status"] = ["good"] * len(tsv_data["component"])
    if "status_description" not in tsv_data:
        tsv_data["status_description"] = ["n/a"] * len(tsv_data["component"])
    for col in ["annotate_method", "annotate_author", "ic_type"]:
        if col not in tsv_data.columns:
            tsv_data[col] = "n/a"

    # load in the tsv file and modify specific columns
    tsv_data.loc[tsv_data["component"] == component, "status"] = status
    tsv_data.loc[tsv_data["component"] == component, "status_description"] = description
    tsv_data.loc[tsv_data["component"] == component, "annotate_method"] = method
    tsv_data.loc[tsv_data["component"] == component, "annotate_author"] = author
    tsv_data.loc[tsv_data["component"] == component, "ic_type"] = label
    tsv_data.to_csv(fname, sep="\t")



# analyze.py
# Old class that I was using when I used freesurfer and MNE-Python to try and compute the PCIlz from the original Casali paper from 2013
import numpy as np
from scipy import stats, signal
from scipy.stats import zscore
import matplotlib.pyplot as plt
import seaborn as sns
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable
from scipy.linalg import svd
from scipy.spatial.distance import pdist, squareform


from typing import Optional, Union, Dict, List, Tuple, Any
import warnings   
import os
from multiprocessing import Pool, cpu_count
from tqdm import tqdm

# MNE imports
import mne
from mne.minimum_norm import (make_inverse_operator, 
                            apply_inverse,
                            write_inverse_operator)
from mne.preprocessing import compute_proj_ecg, compute_proj_eog
from mne import (compute_raw_covariance,
                read_source_spaces,
                setup_source_space,
                make_bem_model,
                make_bem_solution,
                make_forward_solution,
                read_trans,
                read_bem_solution)
from mne.simulation import simulate_evoked, simulate_raw
from mne.viz import plot_alignment, plot_bem


class PCIlz:
    """
    A class for analyzing preprocessed TMS-EEG data.
    
    This class implements various analysis methods including source modeling,
    connectivity analysis, and perturbational complexity index calculation.
    
    Parameters
    ----------
    preprocessed_epochs : mne.Epochs
        The preprocessed TMS-EEG epochs
    subjects_dir : str
        FreeSurfer subjects directory
    subject : str
        Subject name in FreeSurfer directory
        
    Attributes
    ----------
    epochs : mne.Epochs
        The preprocessed epochs
    evoked : mne.Evoked
        The averaged evoked response
    stc : mne.SourceEstimate
        The source estimates (after source modeling)
    """
    
    def __init__(self, 
                 preprocessed_epochs: mne.Epochs,
                 subjects_dir: str,
                 subject: str,
                 trans_path: str):
        
        self.epochs = preprocessed_epochs
        self.evoked = preprocessed_epochs.average()
        self.evoked_sim = None
        self.subjects_dir = subjects_dir
        self.subject = subject
        self.trans_path = trans_path
        
        # Initialize attributes that will be set later
        self.stc = None
        self.forward = None
        self.inverse_operator = None
        self.source_space = None
        self.bem_solution = None
        self.noise_cov = None
        
        # Ensure average reference
        self.evoked.set_eeg_reference('average', projection=True)

    def load_source_space(self, fpath):
       print('Loading source space.. If it does not work there might be file an issue with how paths are passed as strings')
       self.source_space = mne.read_source_spaces(fpath)

    def load_bem_solution(self, fpath):
        print('Loading bem solution.. If it does not work there might be file an issue with how paths are passed as strings')
        self.bem_solution = mne.read_bem_solution(fpath)

    def setup_source_space(self, spacing: str = 'oct6', add_dist: bool = False) -> None:
        """
        Setup the source space for source modeling.
        
        Parameters
        ----------
        spacing : str
            The spacing to use for the source space
        add_dist : bool
            Whether to add distances to the source space
        """
        try:
            self.source_space = mne.setup_source_space(
                subject=self.subject,
                spacing=spacing,
                subjects_dir=self.subjects_dir,
                add_dist=add_dist,
                n_jobs=1
            )
            print("Source space setup complete")
        except Exception as e:
            print(f"Error setting up source space: {str(e)}")
            
    def setup_bem_model(self, ico: Optional[int] = None, 
                       conductivity: Tuple[float, float, float] = (0.3, 0.006, 0.3)) -> None:
        """
        Setup the BEM model for forward modeling.
        
        Parameters
        ----------
        ico : int or None
            The ico parameter for BEM model
        conductivity : tuple
            Conductivity values for the three layers
        """
        try:
            model = mne.make_bem_model(
                subject=self.subject,
                subjects_dir=self.subjects_dir,
                ico=ico,
                conductivity=conductivity
            )
            self.bem_solution = mne.make_bem_solution(model)
            print("BEM model setup complete")
        except Exception as e:
            print(f"Error setting up BEM model: {str(e)}")

    def compute_forward_solution(self, mindist: float = 5.0) -> None:
        """
        Compute the forward solution.
        
        Parameters
        ----------
        mindist : float
            Minimum distance between sources
        """
        if self.source_space is None or self.bem_solution is None:
            raise ValueError("Must setup source space and BEM model first")
            
        try:
            # Get trans file path
            trans_path = self.trans_path
            trans = mne.read_trans(trans_path)
            
            self.forward = mne.make_forward_solution(
                self.evoked.info,
                trans=trans,
                src=self.source_space,
                bem=self.bem_solution,
                mindist=mindist,
                eeg=True,
                meg=False,
                n_jobs=-1
            )
            print("Forward solution computation complete")
        except Exception as e:
            print(f"Error computing forward solution: {str(e)}")

    def compute_inverse_operator(self, loose: float = 1.0, depth: float = 0.8, 
                               fixed: bool = False) -> None:
        """
        Compute the inverse operator.
        
        Parameters
        ----------
        loose : float
            Loose orientation constraint
        depth : float
            Depth weighting
        fixed : bool
            Whether to use fixed orientation
        """
        if self.forward is None:
            raise ValueError("Must compute forward solution first")
            
        # Compute noise covariance from baseline
        self.noise_cov = mne.compute_covariance(
            self.epochs,
            tmin=-0.5,
            tmax=-0.001,
            method='empirical',
            rank='info'
        )
        
        self.inverse_operator = mne.minimum_norm.make_inverse_operator(
            self.evoked.info,
            self.forward,
            self.noise_cov,
            loose=loose,
            depth=depth,
            fixed=fixed
        )
        print("Inverse operator computation complete")

    def compute_source_estimate(self, evoked=True, evoked_sim=False, lambda2: float = 1.0/9.0, 
                              method: str = 'MNE') -> None:
        """
        Compute the source estimate.
        
        Parameters
        ----------
        lambda2 : float
            Regularization parameter
        method : str
            Inverse method to use
        """
        if evoked:
            if self.inverse_operator is None:
                raise ValueError("Must compute inverse operator first")
                
            self.stc = mne.minimum_norm.apply_inverse(
                self.evoked,
                self.inverse_operator,
                lambda2=lambda2,
                method=method,
                pick_ori=None
            )
            print("Source estimate computation complete")

        elif evoked_sim:
            if self.inverse_operator is None:
                raise ValueError("Must compute inverse operator first")
                
            self.stc = mne.minimum_norm.apply_inverse(
                self.evoked_sim,
                self.inverse_operator,
                lambda2=lambda2,
                method=method,
                pick_ori=None
            )
            print("Source estimate computation complete using simulated evoked object")



    def detect_significant_sources(self, n_bootstraps: int = 500, 
                                alpha: float = 0.01,
                                baseline: Tuple[float, float] = (-0.5, -0.001)) -> np.ndarray:
        """
        Detect significant TMS-evoked cortical activation following Casali et al. 2013.
        """
        if self.inverse_operator is None:
            raise ValueError("Must compute inverse operator first")
            
        # Get source estimates for all trials
        stcs = []
        for trial in self.epochs:
            stc = mne.minimum_norm.apply_inverse(trial, self.inverse_operator,
                                                lambda2=1.0/9.0, method='MNE')
            stcs.append(stc.data)
        source_data = np.stack(stcs)  # Shape: (trials, sources, times)
        
        # Get baseline data
        baseline_mask = ((stc.times >= baseline[0]) & (stc.times <= baseline[1]))
        baseline_data = source_data[:, :, baseline_mask]
        
        # Calculate baseline mean and std for each source
        baseline_mean = np.mean(np.mean(baseline_data, axis=0), axis=1)[:, np.newaxis]
        baseline_std = np.std(np.mean(baseline_data, axis=0), axis=1)[:, np.newaxis]
        
        # Normalize source data
        normalized_data = ((np.mean(source_data, axis=0) - baseline_mean) / baseline_std)
        
        # Run bootstrap procedure
        max_abs_bootstraps = []
        for _ in range(n_bootstraps):
            # Shuffle baseline data at single-trial level for each source
            surrogate_data = np.stack([np.stack([np.random.permutation(bd) for bd in baseline_data[:, s]]) 
                                    for s in range(baseline_data.shape[1])])
            
            # Calculate mean and normalize
            surrogate_mean = np.mean(surrogate_data, axis=0)
            normalized_surrogate = (surrogate_mean - baseline_mean) / baseline_std
            
            # Store maximum absolute value across sources and times
            max_abs_bootstraps.append(np.max(np.abs(normalized_surrogate)))
            
        # Calculate significance threshold
        threshold = np.percentile(max_abs_bootstraps, (1 - alpha) * 100)
        
        # Create binary significance matrix
        SS = (np.abs(normalized_data) > threshold).astype(int)
        
        return SS

    
    def calculate_pci(self, SS: np.ndarray, min_duration_ms: int = 10) -> float:
        """
        Calculate Perturbational Complexity Index from significant sources matrix.
        
        Parameters
        ----------
        SS : np.ndarray
            Binary matrix of significant sources
        min_duration_ms : int
            Minimum duration for temporal compression
            
        Returns
        -------
        float
            PCI value
        """
        # Check data validity
        p_1 = np.mean(SS)
        if p_1 < 0.01 or p_1 > 0.99:
            print(f"Warning: Proportion of significant values ({p_1:.3f}) outside valid range")
            return 0
        print("Calculating source entropy.. ")   
        # Calculate source entropy
        H_L = -p_1 * np.log2(p_1) - (1-p_1) * np.log2(1-p_1)
        if H_L < 0.08:
            print(f"Warning: Initial entropy ({H_L:.3f}) too low")
            return 0
        print(f"Source entropy: {H_L}")    
        # Convert to binary sequence and calculate complexity
        binary_sequence = SS.flatten(order='F')
        c_L = self._custom_lempel_ziv_complexity(binary_sequence)
        print("Calculating PCI..")
        # Calculate final PCI
        L = len(binary_sequence)
        pci = (c_L * np.log2(L)) / (L * H_L)
        print(f"PCI: {pci}")
        return pci
        
    @staticmethod
    def _custom_lempel_ziv_complexity(binary_sequence: np.ndarray) -> int:
        """
        Calculate Lempel-Ziv complexity of a binary sequence.
        
        Parameters
        ----------
        binary_sequence : np.ndarray
            Binary sequence to analyze
            
        Returns
        -------
        int
            Lempel-Ziv complexity value
        """
        binary_str = ''.join(binary_sequence.astype(str))
        n = len(binary_str)
        
        complexity = 1
        current_position = 0
        current_length = 1
        patterns = {binary_str[0]: True}
        
        while current_position + current_length <= n:
            pattern = binary_str[current_position:current_position + current_length]
            
            if pattern in patterns:
                current_length += 1
            else:
                patterns[pattern] = True
                complexity += 1
                current_position += current_length
                current_length = 1
                
                if current_position == n:
                    break
                    
        return complexity

    def plot_source_activations(self, timepoints: List[float], 
                              save_path: Optional[str] = None) -> None:
        """
        Plot source activations at specific timepoints.
        
        Parameters
        ----------
        timepoints : list
            List of timepoints (in seconds) to plot
        save_path : str, optional
            Path to save the plots
        """
        if self.stc is None:
            raise ValueError("Must compute source estimate first")
            
        # Find closest times in data
        time_indices = [np.abs(self.stc.times - tp).argmin() for tp in timepoints]
        actual_times = [self.stc.times[idx] for idx in time_indices]
        
        # Visualization parameters
        surfer_kwargs = dict(
            subject=self.subject,
            hemi='split',
            subjects_dir=self.subjects_dir,
            views=['lat', 'med'],
            clim=dict(kind='percent', pos_lims=[80, 90, 95]),
            colormap='mne',
            background='white',
            time_viewer=False,
            smoothing_steps=10,
            size=(1000, 400)
        )
        
        # Plot each timepoint
        brains = []
        for actual_time in actual_times:
            brain = self.stc.plot(
                initial_time=actual_time,
                **surfer_kwargs
            )
            brain.add_text(0.1, 0.9, f'Time: {actual_time*1000:.0f} ms',
                         font_size=14)
            brains.append(brain)
            
            if save_path:
                brain.save_image(f"{save_path}_time_{int(actual_time*1000)}ms.png")
                
        return brains
    
    def simulate_eeg(self, evoked=False, simple=False, nave=30, 
                    iir_filter=None, random_state=None, verbose=True,
                    pick_ori='normal'):
        """
        Simulate EEG data using either raw or evoked simulation.
        
        Parameters
        ----------
        evoked : bool
            If True, simulates evoked data. If False, simulates raw data.
        simple : bool
            Whether to only use epochs.info for simulation or to use the source estimate 
            and forward solution
        nave : int
            Number of averaged epochs (only for evoked simulation)
        iir_filter : None | array
            IIR filter coefficients (denominator) e.g. [1, -1, 0.2] (only for evoked simulation)
        random_state : None | int | instance of RandomState
            Random seed for reproducibility
        verbose : bool
            Controls the logging of mne simulation functions
        pick_ori : str
            The source orientation to use. Options:
            - 'normal' : Orient sources normal to cortical surface
            - None : Use magnitude of source currents
            - 'vector' : Keep vectorial source directions

        Returns
        -------
        raw_sim : Raw object or Evoked object
            The simulated EEG data
        """
        info = self.epochs.info
        
        # If noise covariance isn't computed yet, compute it
        if self.noise_cov is None:
            print("Computing noise covariance from baseline period...")
            self.noise_cov = mne.compute_covariance(
                self.epochs,
                tmin=-0.5,
                tmax=-0.001,
                method='empirical',
                rank='info'
            )
        
        if evoked:
            if not simple:
                    try:
                        self.evoked_sim = simulate_evoked(
                            fwd=self.forward,
                            stc=self.stc,
                            info=info,
                            cov=self.noise_cov,  # Using class noise covariance
                            nave=nave,
                            iir_filter=iir_filter,
                            random_state=random_state,
                            use_cps=True,
                            verbose=verbose
                        )
                        
                        # Add projectors back if needed
                        if len(info['projs']) > 0:
                            self.evoked_sim.add_proj(info['projs'])
                            
                        return self.evoked_sim
                        
                    except Exception as e:
                        print(f"Error during evoked simulation: {str(e)}")
                        return None
            else:
                print("Simple simulation not available for evoked data. Please compute forward solution and source estimate.")
                return None
        else:
            if not simple:
                    try:
                        raw_sim = simulate_raw(
                            info=info, 
                            stc=self.stc, 
                            forward=self.forward, 
                            verbose=verbose
                        )
                        return raw_sim
                    except Exception as e:
                        print(f"Error during raw simulation: {str(e)}")
                        return None
            else:
                raw_sim = simulate_raw(info=info, verbose=verbose)
                return raw_sim



    def plot_tms_eeg_analysis(self, SS: np.ndarray, pci: float, 
                             title: Optional[str] = None,
                             save_path: Optional[str] = None,
                             color_palette: Optional[List[str]] = None,
                            evoked_ylim: List[float] = [-7, 7],
                            evoked_xlim: Optional[Tuple[float, float]] = (-0.1, 0.5)) -> None:
        
        """
        Create a visualization of TMS-EEG analysis results.
        
        Parameters
        ----------
        SS : np.ndarray
            Binary matrix of significant sources
        pci : float
            Calculated PCI value
        title : str, optional
            Main title for the entire figure
        save_path : str, optional
            Path to save the figure
        color_palette : list of str, optional
            Custom color palette for plots. If None, uses default seaborn deep palette
        evoked_ylim : list of float
            Y-axis limits for evoked plot [lower, upper]
        evoked_xlim : tuple of float, optional
            X-axis limits for evoked plot in seconds (start_time, end_time)
        """
        # Set color parameters
        if color_palette is None:
            colors = sns.color_palette("deep")
        else:
            colors = color_palette
            
        # Create figure
        fig = plt.figure(figsize=(16, 14))
        fig.patch.set_facecolor('white')
        gs = plt.GridSpec(4, 2, height_ratios=[1, 1, 1, 0.8])
        
        if title:
            fig.suptitle(title, fontsize=16, fontweight='bold', y=0.98)
        
        # 1. Global Field Power
        ax_gfp = fig.add_subplot(gs[0, :])
        ax_gfp.set_facecolor('#f8f8f8')
        ax_gfp.grid(True, linestyle='--', alpha=0.3, color='gray')
        
        gfp = np.std(self.stc.data, axis=0)
        times = self.stc.times * 1000  # Convert to ms
        
        # Plot GFP with confidence interval and styling
        ax_gfp.plot(times, gfp, color=colors[0], linewidth=2, label='GFP')
        ax_gfp.fill_between(times, 
                        gfp-stats.sem(self.stc.data, axis=0),
                        gfp+stats.sem(self.stc.data, axis=0),
                        color=colors[0], alpha=0.2)
        
        # Add component markers with enhanced styling
        components = {
            'P30': (30, colors[1]), 
            'N45': (45, colors[2]), 
            'P55': (55, colors[3]),
            'N100': (100, colors[4]), 
            'P180': (180, colors[5])
        }
        
        max_gfp = np.max(gfp)
        for comp, (t, color) in components.items():
            if t > times[0] and t < times[-1]:
                ax_gfp.axvline(x=t, color=color, linestyle=':', alpha=0.7)
                ax_gfp.text(t, max_gfp*1.1, comp, rotation=90,
                        horizontalalignment='center', 
                        color=color, fontweight='bold')
        
        # Add TMS pulse marker
        ax_gfp.axvline(x=0, color='red', linestyle='--', linewidth=2, label='TMS pulse')
        
        ax_gfp.set_title('Global Field Power Over Time', fontsize=12, fontweight='bold', pad=15)
        ax_gfp.set_xlabel('Time (ms)', fontsize=10)
        ax_gfp.set_ylabel('Global Field Power (μV)', fontsize=10)
        ax_gfp.legend(loc='upper right', frameon=True, fancybox=True, shadow=True)
        
        # 2. Butterfly Plot with enhanced styling
        ax_butterfly = fig.add_subplot(gs[1, :])
        ax_butterfly.set_facecolor('#f8f8f8')
        ax_butterfly.grid(True, linestyle='--', alpha=0.3, color='gray')
        
        self.evoked.plot(axes=ax_butterfly, show=False, spatial_colors=True,
                        ylim=dict(eeg=evoked_ylim), 
                        xlim=evoked_xlim,  # Changed this line
                        titles=dict(eeg='Averaged Evoked Response'))
        
        # Add component markers to butterfly plot
        max_val = max(evoked_ylim)
        for comp, (t, color) in components.items():
            if t > times[0] and t < times[-1]:
                ax_butterfly.axvline(x=t/1000, color=color, linestyle=':', alpha=0.7)
                ax_butterfly.text(t/1000, max_val*1.1, comp, rotation=90,
                                horizontalalignment='center', 
                                color=color, fontweight='bold')
        
        # 3. Significant Sources Heatmap with enhanced styling
        ax_ss = fig.add_subplot(gs[2, :])
        ax_ss.set_facecolor('#f8f8f8')
        ax_ss.grid(True, linestyle='--', alpha=0.3, color='gray')

        im = ax_ss.imshow(SS, aspect='auto', cmap='RdBu_r',
                        extent=[0, 300, 0, SS.shape[0]])

        # Add percentage of significant sources with styled twin axis
        ax_ss2 = ax_ss.twinx()
        sig_percent = np.mean(SS, axis=0) * 100

        # Fix: Properly mask the time points to match significant sources
        time_mask = (times >= 0) & (times <= 300)
        plot_times = times[time_mask]
        if len(plot_times) > len(sig_percent):
            plot_times = plot_times[:len(sig_percent)]
        elif len(plot_times) < len(sig_percent):
            sig_percent = sig_percent[:len(plot_times)]

        ax_ss2.plot(plot_times, sig_percent, 
                    color=colors[2], linewidth=2, alpha=0.7)
        ax_ss2.set_ylabel('% Significant Sources', color=colors[2], 
                        fontweight='bold', fontsize=10)
        ax_ss2.tick_params(axis='y', colors=colors[2])
        
        ax_ss.set_xlabel('Time (ms)', fontsize=10)
        ax_ss.set_ylabel('Source Index', fontsize=10)
        ax_ss.set_title('Significant Sources Over Time', 
                        fontsize=12, fontweight='bold', pad=15)
        
        # Add colorbar with styling
        divider = make_axes_locatable(ax_ss)
        cax = divider.append_axes("right", size="2%", pad=0.05)
        plt.colorbar(im, cax=cax, label='Significance')
        
        # 4. Analysis Statistics with enhanced styling
        ax_stats = fig.add_subplot(gs[3, 0])
        ax_stats.set_facecolor('#f8f8f8')
        ax_stats.axis('off')
        
        stats_text = f"""
        TMS-EEG Analysis Metrics:
        -------------------------
        
        PCI Value: {pci:.3f}
        Total Significant Sources: {np.sum(SS):,}
        Peak Time: {times[np.argmax(np.mean(SS, axis=0))]: .1f} ms
        Duration: {np.sum(np.any(SS, axis=0)) * (times[1]-times[0]): .1f} ms
        
        Coverage Statistics:
        ------------------
        Total Sources: {SS.shape[0]:,}
        Time Points: {SS.shape[1]:,}
        Significance Rate: {100 * np.sum(SS) / SS.size:.2f}%
        """
        ax_stats.text(0.05, 0.95, stats_text, 
                    fontsize=11, va='top', 
                    family='monospace',
                    bbox=dict(facecolor='white', 
                            alpha=0.8,
                            edgecolor='gray',
                            boxstyle='round,pad=1'))
        
        # 5. Source Distribution with enhanced styling
        ax_hist = fig.add_subplot(gs[3, 1])
        ax_hist.set_facecolor('#f8f8f8')
        ax_hist.grid(True, linestyle='--', alpha=0.3, color='gray')
        
        peak_time_idx = np.argmax(np.std(self.stc.data, axis=0))
        peak_data = self.stc.data[:, peak_time_idx]
        
        sns.histplot(data=peak_data, ax=ax_hist, 
                    bins=50, color=colors[0], 
                    alpha=0.7, stat='density')
        
        ax_hist.set_xlabel('Source Amplitude', fontsize=10)
        ax_hist.set_ylabel('Density', fontsize=10)
        ax_hist.set_title(f'Source Distribution at Peak ({times[peak_time_idx]:.1f} ms)',
                        fontsize=12, fontweight='bold', pad=15)
        
        # Adjust layout
        plt.tight_layout()
        
        if save_path:
            plt.savefig(save_path, dpi=300, bbox_inches='tight')
            
        return fig
    
    



from __future__ import annotations  # c.f. PEP 563, PEP 649

from typing import TYPE_CHECKING

import numpy as np
from mne import BaseEpochs
from mne.io import BaseRaw
from mne.utils import warn
from numpy.typing import NDArray
from scipy.signal import resample_poly

from ..utils._checks import _validate_inst_and_ica
from ._utils import _gdatav4, _mne_to_eeglab_locs, _next_power_of_2, _pol2cart

if TYPE_CHECKING:
    from typing import Union

    from mne.preprocessing import ICA


def get_iclabel_features(inst: Union[BaseRaw, BaseEpochs], ica: ICA):
    """Generate the features for ICLabel neural network.

    Parameters
    ----------
    inst : Raw | Epochs
        MNE Raw/Epoch instance with data array in Volts.
    ica : ICA
        MNE ICA decomposition.

    Returns
    -------
    topo : array of shape (32, 32, 1, n_components)
        The topoplot feature.
    psd : array of shape (1, 100, 1, n_components)
        The psd feature.
    autocorr : array of shape (1, 100, 1, n_components)
        The autocorrelations feature. Depending on the length of the
        raw data passed in, different methods of computing autocorrelation
        will be used. See :footcite:t:`PionTonachini2019` for details.

    References
    ----------
    .. footbibliography::
    """
    _validate_inst_and_ica(inst, ica)
    if "eeg" not in inst:
        raise RuntimeError(
            "Could not find EEG channels in the provided "
            f"{'Raw' if isinstance(inst, BaseRaw) else 'Epochs'} instance. The ICLabel "
            "model was fitted on EEG data and is not suited for other types of "
            "channels."
        )

    # TODO: 'custom_ref_applied' does not necessarily correspond to a CAR reference.
    # At the moment, the reference of the EEG data is not stored in the info.
    # c.f. https://github.com/mne-tools/mne-python/issues/8962
    if inst.info["custom_ref_applied"] == 0:
        warn(
            f"The provided {'Raw' if isinstance(inst, BaseRaw) else 'Epochs'} instance "
            "does not seem to be referenced to a common average reference (CAR). "
            "ICLabel was designed to classify features extracted from an EEG dataset "
            "referenced to a CAR (see the 'set_eeg_reference()' method for Raw and "
            "Epochs instances)."
        )
    if inst.info["highpass"] != 1 or inst.info["lowpass"] != 100:
        warn(
            f"The provided {'Raw' if isinstance(inst, BaseRaw) else 'Epochs'} instance "
            "is not filtered between 1 and 100 Hz. "
            "ICLabel was designed to classify features extracted from an EEG dataset "
            "bandpass filtered between 1 and 100 Hz (see the 'filter()' method for Raw "
            "and Epochs instances)."
        )
    # confirm that the ICA uses an infomax extended
    method_ = ica.method not in ("infomax", "picard")
    extended_ = "extended" not in ica.fit_params or ica.fit_params["extended"] is False
    ortho_ = "ortho" not in ica.fit_params or ica.fit_params["ortho"] is True
    ortho_ = ortho_ if ica.method == "picard" else False
    if any((method_, extended_, ortho_)):
        warn(
            f"The provided ICA instance was fitted with a '{ica.method}' algorithm. "
            "ICLabel was designed with extended infomax ICA decompositions. To use the "
            "extended infomax algorithm, use the 'mne.preprocessing.ICA' instance with "
            "the arguments 'ICA(method='infomax', fit_params=dict(extended=True))' "
            "(scikit-learn) or 'ICA(method='picard', fit_params=dict(ortho=False, "
            "extended=True))' (python-picard)."
        )

    icawinv, _ = _retrieve_eeglab_icawinv(ica)
    icaact = _compute_ica_activations(inst, ica)

    # compute topographic feature (float32)
    topo = _eeg_topoplot(inst, icawinv, ica.ch_names)

    # compute psd feature (float32)
    psd = _eeg_rpsd(inst, ica, icaact)

    # compute autocorr feature (float32)
    if isinstance(inst, BaseRaw):
        if 5 < inst.times.size / inst.info["sfreq"]:
            autocorr = _eeg_autocorr_welch(inst, ica, icaact)
        else:
            autocorr = _eeg_autocorr(inst, ica, icaact)
    else:
        autocorr = _eeg_autocorr_fftw(inst, ica, icaact)

    # scale by 0.99
    topo *= 0.99
    psd *= 0.99
    autocorr *= 0.99

    return topo, psd, autocorr


def _retrieve_eeglab_icawinv(
    ica: ICA,
) -> tuple[NDArray[float], NDArray[float]]:
    """
    Retrieve 'icawinv' from an MNE ICA instance.

    Parameters
    ----------
    ica : ICA
        MNE ICA decomposition.

    Returns
    -------
    icawinv : array
    weights : array
    """
    n_components = ica.n_components_
    s = np.sqrt(ica.pca_explained_variance_)[:n_components]
    u = ica.unmixing_matrix_ / s
    v = ica.pca_components_[:n_components, :]
    weights = (u * s) @ v
    icawinv = np.linalg.pinv(weights)
    # sanity-check
    assert icawinv.shape[-1] == ica.n_components_
    assert icawinv.ndim == 2
    return icawinv, weights


def _compute_ica_activations(
    inst: Union[BaseRaw, BaseEpochs], ica: ICA
) -> NDArray[float]:
    """Compute the ICA activations 'icaact' variable from an MNE ICA instance.

    Parameters
    ----------
    inst : Raw | Epoch
        MNE Raw/Epoch instance with data array in Volts.
    ica : ICA
        MNE ICA decomposition.

    Returns
    -------
    icaact : array
        raw: (n_components, n_samples)
        epoch: (n_components, n_samples, n_trials)

    Notes
    -----
    In EEGLAB, the ICA activations are computed after the data and the ICA
    decomposition are re-referenced to a common average, if the field 'EEG.ref'
    is different from 'averef'. The EEGLAB sample dataset's field 'EEG.ref' is
    set to 'common', thus triggering the re-referencing with 'pop_reref' which
    seems to be buggy and breaks the ICA solution. After 'pop_reref' is called,
    the relation 'inv(EEG.icaweights * EEG.icasphere) = EEG.icawinv' is not
    respected anymore.

    Additionally, 'pop_reref' changes the field 'EEG.ref' to 'average'. It is
    assumed that 'common', 'average' and 'averef' are all denoting a common
    average reference.
    """
    icawinv, weights = _retrieve_eeglab_icawinv(ica)
    icasphere = np.eye(icawinv.shape[0])
    data = inst.get_data(picks=ica.ch_names) * 1e6
    icaact = (weights[0 : ica.n_components_, :] @ icasphere) @ data
    # move trial (epoch) dimension to the end
    if icaact.ndim == 3:
        assert isinstance(inst, BaseEpochs)  # sanity-check
        icaact = icaact.transpose([1, 2, 0])
    return icaact


# ----------------------------------------------------------------------------
def _eeg_topoplot(
    inst: Union[BaseRaw, BaseEpochs], icawinv: NDArray[float], picks: list[str]
) -> NDArray[float]:
    """Topoplot feature."""
    ncomp = icawinv.shape[-1]
    topo = np.zeros((32, 32, 1, ncomp))
    rd, th = _mne_to_eeglab_locs(inst, picks)
    th = np.pi / 180 * th  # convert degrees to radians
    for it in range(ncomp):
        temp_topo = _topoplotFast(icawinv[:, it], rd, th)
        np.nan_to_num(temp_topo, copy=False)  # set NaN values to 0 in-place
        topo[:, :, 0, it] = temp_topo / np.max(np.abs(temp_topo))
    return topo.astype(np.float32)


def _topoplotFast(
    values: NDArray[float], rd: NDArray[float], th: NDArray[float]
) -> NDArray[float]:
    """Implement topoplotFast.m from MATLAB. Each topographic map is a 32x32 images."""
    # constants
    GRID_SCALE = 32  # number of pixels
    rmax = 0.5  # actual head radius

    # convert electrode locations from polar to cartesian coordinates
    x, y = _pol2cart(th, rd)

    # prepare coordinates
    # Comments in MATLAB (L750:753) are:
    #   default: just outside the outermost electrode location
    #   default: plot out to the 0.5 head boundary
    #   don't plot channels with Rd > 1 (below head)
    plotrad = min(1, np.max(rd) * 1.02)
    plotrad = max(plotrad, 0.5)

    # TODO: Selection of channels.
    # For interpolation, only the channels inside the interpolation square are
    # considered. c.f. L839:843.

    # Squeeze channel location to <= rmax
    # Comments in MATLAB (L894:908)
    #   squeeze electrode arc_lengths towards the vertex to plot all inside the
    #   head cartoon
    squeezefac = rmax / plotrad
    rd *= squeezefac
    x *= squeezefac
    y *= squeezefac
    # convert to float63
    x = x.astype(np.float64)
    y = y.astype(np.float64)

    # Find limits for interpolation
    xmin = min(-rmax, np.min(x))
    xmax = max(rmax, np.max(x))
    ymin = min(-rmax, np.min(y))
    ymax = max(rmax, np.max(y))

    # Interpolate scalp map data
    xi = np.linspace(xmin, xmax, GRID_SCALE).astype(np.float64).reshape((1, -1))
    yi = np.linspace(ymin, ymax, GRID_SCALE).astype(np.float64).reshape((1, -1))
    # additional step for gdatav4 compared to MATLAB: linspace to meshgrid
    XQ, YQ = np.meshgrid(xi, yi)
    Xi, Yi, Zi = _gdatav4(x, y, values.reshape((-1, 1)), XQ, YQ)
    # additional step for gdatav4 compared to MATLAB: transpose
    Zi = Zi.T

    # Mask out data outside the head
    mask = np.sqrt(np.power(Xi, 2) + np.power(Yi, 2)) <= rmax
    Zi[~mask] = np.nan

    return Zi


# ----------------------------------------------------------------------------
def _eeg_rpsd(
    inst: Union[BaseRaw, BaseEpochs], ica: ICA, icaact: NDArray[float]
) -> NDArray[float]:
    """PSD feature."""
    assert isinstance(inst, (BaseRaw, BaseEpochs))  # sanity-check
    constants = _eeg_rpsd_constants(inst, ica)
    psd = _eeg_rpsd_compute_psdmed(inst, icaact, *constants)
    psd = _eeg_rpsd_format(psd)
    return psd


def _eeg_rpsd_constants(
    inst: Union[BaseRaw, BaseEpochs],
    ica: ICA,
) -> tuple[int, int, int, int, NDArray[int], NDArray[float], NDArray[int]]:
    """Compute the constants before ``randperm`` is used to compute the subset."""
    # in MATLAB, 'pct_data' variable is never provided and is always initialized
    # to 100. 'pct_data' is only used in a division by 100.. and thus has no
    # impact and is omitted here.
    # in MATLAB, 'nfreqs' variable is always provided as 100 to this function,
    # thus it is either equal to 100 or to the nyquist frequency depending on
    # the nyquist frequency.

    nyquist = np.floor(inst.info["sfreq"] / 2).astype(int)
    nfreqs = nyquist if nyquist < 100 else 100

    ncomp = ica.n_components_
    n_points = min(inst.times.size, int(inst.info["sfreq"]))
    window = np.hamming(n_points)
    cutoff = np.floor(inst.times.size / n_points) * n_points

    # python is 0-index while matlab is 1-index, thus (1:n_points) becomes
    # np.arange(0, n_points) since 'index' is used to select from arrays.
    range_ = np.ceil(np.arange(0, cutoff - n_points + n_points / 2, n_points / 2))
    index = np.tile(range_, (n_points, 1)).T + np.arange(0, n_points)
    index = index.T.astype(int)

    # different behaviors based on EEG.trials, i.e. raw or epoch
    if isinstance(inst, BaseRaw):
        n_seg = index.shape[1]
    if isinstance(inst, BaseEpochs):
        n_seg = index.shape[1] * len(inst)

    # in MATLAB: 'subset = randperm(n_seg, ceil(n_seg * pct_data / 100));'
    # which is basically: 'subset = randperm(n_seg, n_seg);'
    # np.random.seed() can be used to fix the seed to the same value as MATLAB,
    # but since the 'randperm' equivalent in numpy does not exist, it is not
    # possible to reproduce the output in python.
    # 'subset' is used to select from arrays and is 0-index in Python while its
    # 1-index in MATLAB.
    rng = np.random.default_rng()
    subset = rng.permutation(range(n_seg))  # 0-index

    return ncomp, nfreqs, n_points, nyquist, index, window, subset


def _eeg_rpsd_compute_psdmed(
    inst: Union[BaseRaw, BaseEpochs],
    icaact: NDArray[float],
    ncomp: int,
    nfreqs: int,
    n_points: int,
    nyquist: int,
    index: NDArray[int],
    window: NDArray[float],
    subset: NDArray[int],
) -> NDArray[float]:
    """Compute the variable 'psdmed', annotated as windowed spectrums."""
    denominator = inst.info["sfreq"] * np.sum(np.power(window, 2))
    psdmed = np.zeros((ncomp, nfreqs))
    for it in range(ncomp):
        # Compared to MATLAB, shapes differ as the component dimension (size 1)
        # was squeezed.
        if isinstance(inst, BaseRaw):
            temp = np.hstack([icaact[it, index[:, k]] for k in range(index.shape[-1])])
            temp = temp.reshape(*index.shape, order="F")
            # equivalent to:
            # np.vstack([icaact[it, index[:, k]] for k in range(index.shape[-1])]).T
        elif isinstance(inst, BaseEpochs):
            temp = np.vstack(
                [icaact[it, index[:, k], :] for k in range(index.shape[-1])]
            )
            temp = temp.reshape(index.shape[0], index.shape[1] * len(inst), order="F")
        temp = (temp[:, subset].T * window).T
        temp = np.fft.fft(temp, n_points, axis=0)
        temp = temp * np.conjugate(temp)
        temp = temp[1 : nfreqs + 1, :] * 2 / denominator
        if nfreqs == nyquist:
            temp[-1, :] = temp[-1, :] / 2
        psdmed[it, :] = 20 * np.real(np.log10(np.median(temp, axis=1)))

    return psdmed


def _eeg_rpsd_format(
    psd: NDArray[float],
) -> NDArray[float]:
    """Apply the formatting steps after 'eeg_rpsd.m'."""
    # extrapolate or prune as needed
    nfreq = psd.shape[1]
    if nfreq < 100:
        psd = np.concatenate([psd, np.tile(psd[:, -1:], (1, 100 - nfreq))], axis=1)

    # undo notch filter
    for linenoise_ind in (50, 60):
        # 'linenoise_ind' is used for array selection in psd, which is
        # 0-index in Python and 1-index in MATLAB.
        linenoise_ind -= 1
        linenoise_around = np.array([linenoise_ind - 1, linenoise_ind + 1])
        # 'linenoise_around' is used for array selection in psd, which is
        # 0-index in Python and 1-index in MATLAB.
        difference = (psd[:, linenoise_around].T - psd[:, linenoise_ind]).T
        notch_ind = np.all(5 < difference, axis=1)
        if any(notch_ind):
            # Numpy doesn't like the selection '[notch_ind, linenoise_ind]' with
            # 'notch_ind' as a bool mask. 'notch_ind' is first converted to int.
            # Numpy doesn't like the selection '[notch_ind, linenoise_around]'
            # with both defined as multi-values 1D arrays (or list). To get
            # around, the syntax [notch_ind[:, None], linenoise_around] is used.
            # That syntax works only with arrays (not list).
            notch_ind = np.where(notch_ind)[0]
            psd[notch_ind, linenoise_ind] = np.mean(
                psd[notch_ind[:, None], linenoise_around], axis=-1
            )

    # normalize
    psd = np.divide(psd.T, np.max(np.abs(psd), axis=-1)).T
    # reshape and cast
    return psd[:, :, np.newaxis, np.newaxis].transpose([2, 1, 3, 0]).astype(np.float32)


def _eeg_autocorr_welch(
    raw: BaseRaw, ica: ICA, icaact: NDArray[float]
) -> NDArray[float]:
    """Autocorrelation feature applied on raw object with at least 5 * fs samples.

    MATLAB: 'eeg_autocorr_welch.m'.
    """
    assert isinstance(raw, BaseRaw)  # sanity-check

    # in MATLAB, 'pct_data' variable is never provided and is always initialized
    # to 100. 'pct_data' is only used in an 'if' statement reached if 'pct_data'
    # is different than 100.. thus, 'pct_data' is not used by this autocorrelation
    # function and is omitted here.

    # setup constants
    ncomp = ica.n_components_
    n_points = min(raw.times.size, int(raw.info["sfreq"] * 3))
    nfft = _next_power_of_2(2 * n_points - 1)
    cutoff = np.floor(raw.times.size / n_points) * n_points
    range_ = np.ceil(np.arange(0, cutoff - n_points + n_points / 2, n_points / 2))
    index = np.tile(range_, (n_points, 1)).T + np.arange(0, n_points)
    # python uses 0-index and matlab uses 1-index
    # python is 0-index while matlab is 1-index, thus (1:n_points) becomes
    # np.arange(0, n_points) since 'index' is used to select from arrays.
    index = index.T.astype(int)

    # separate data segments
    temp = np.hstack([icaact[:, index[:, k]] for k in range(index.shape[-1])])
    segments = temp.reshape(ncomp, *index.shape, order="F")

    """
    # Just in case, here is the 'if' statement when 'pct_data' is different
    # than 100.

    n_seg = index.shape[1]
    # In MATLAB: n_seg = size(index, 2) * EEG.trials;
    # However, this function is only called on RAW dataset with EEG.trials
    # equal to 1.

    # in MATLAB: 'subset = randperm(n_seg, ceil(n_seg * pct_data / 100));'
    # which is basically: 'subset = randperm(n_seg, n_seg);'
    # np.random.seed() can be used to fix the seed to the same value as MATLAB,
    # but since the 'randperm' equivalent in numpy does not exist, it is not
    # possible to reproduce the output in python.
    # 'subset' is used to select from arrays and is 0-index in Python while its
    # 1-index in MATLAB.
    subset = np.random.permutation(range(n_seg))  # 0-index
    temp = np.hstack([icaact[:, index[:, k]] for k in range(index.shape[-1])])
    temp = temp.reshape(ncomp, *index.shape, order='F')
    segments = temp[:, :, subset]
    """

    # calc autocorrelation
    ac = np.zeros((ncomp, nfft))
    for it in range(ncomp):
        x = np.fft.fft(segments[it, :, :], nfft, axis=0)
        ac[it, :] = np.mean(np.power(np.abs(x), 2), axis=1)
    ac = np.fft.ifft(ac)

    # normalize
    # In MATLAB, 2 scenarios are defined:
    # - EEG.pnts < EEG.srate, which never occurs since then raw provided to
    # this autocorrelation function last at least 5 second.
    # - EEG.pnts > EEG.srate, implemented below.
    ac = ac[:, : int(raw.info["sfreq"]) + 1]
    # build the (3-line!) denominator
    arr1 = np.arange(n_points, n_points - int(raw.info["sfreq"]), -1)
    arr1 = np.hstack([arr1, [np.max([1, n_points - int(raw.info["sfreq"])])]])
    den = np.tile(ac[:, 0], (arr1.size, 1))
    den = den.T * arr1 / n_points
    # finally..
    ac = np.divide(ac, den)

    # resample to 1 second at 100 samples/sec
    resamp = _resample(ac, raw.info["sfreq"])
    resamp = resamp[:, 1:, np.newaxis, np.newaxis].transpose([2, 1, 3, 0])
    return np.real(resamp).astype(np.float32)


def _eeg_autocorr(raw: BaseRaw, ica: ICA, icaact: NDArray[float]) -> NDArray[float]:
    """Autocorr applied on raw object without enough samples for eeg_autocorr_welch.

    MATLAB: 'eeg_autocorr.m'.
    """
    assert isinstance(raw, BaseRaw)  # sanity-check

    # in MATLAB, 'pct_data' variable is neither provided or used, thus it is
    # omitted here.
    ncomp = ica.n_components_
    nfft = _next_power_of_2(2 * raw.times.size - 1)

    c = np.zeros((ncomp, nfft))
    for it in range(ncomp):
        # in MATLAB, 'mean' does nothing here. It looks like it was included
        # for a case where epochs are provided, which never happens with this
        # autocorrelation function.
        x = np.power(np.abs(np.fft.fft(icaact[it, :], n=nfft)), 2)
        c[it, :] = np.real(np.fft.ifft(x))

    if raw.times.size < raw.info["sfreq"]:
        zeros = np.zeros((c.shape[0], int(raw.info["sfreq"]) - raw.times.size + 1))
        ac = np.hstack([c[:, : raw.times.size], zeros])
    else:
        ac = c[:, : int(raw.info["sfreq"]) + 1]

    # normalize by 0-tap autocorrelation
    ac = np.divide(ac.T, ac[:, 0]).T

    # resample to 1 second at 100 samples/sec
    resamp = _resample(ac, raw.info["sfreq"])
    resamp = resamp[:, 1:, np.newaxis, np.newaxis].transpose([2, 1, 3, 0])
    return resamp.astype(np.float32)


def _eeg_autocorr_fftw(
    epochs: BaseEpochs, ica: ICA, icaact: NDArray[float]
) -> NDArray[float]:
    """Autocorrelation feature applied on epoch object.

    MATLAB: 'eeg_autocorr_fftw.m'.
    """
    assert isinstance(epochs, BaseEpochs)  # sanity-check

    # in MATLAB, 'pct_data' variable is neither provided or used, thus it is
    # omitted here.
    ncomp = ica.n_components_
    nfft = _next_power_of_2(2 * epochs.times.size - 1)

    ac = np.zeros((ncomp, nfft))
    for it in range(ncomp):
        x = np.fft.fft(icaact[it, :, :], nfft, axis=0)
        ac[it, :] = np.mean(np.power(np.abs(x), 2), axis=1)

    ac = np.fft.ifft(ac)

    if epochs.times.size < epochs.info["sfreq"]:
        zeros = np.zeros(
            (ac.shape[0], int(epochs.info["sfreq"]) - epochs.times.size + 1)
        )
        ac = np.hstack([ac[:, : epochs.times.size], zeros])
    else:
        ac = ac[:, : int(epochs.info["sfreq"]) + 1]

    ac = np.divide(ac.T, ac[:, 0]).T

    # resample to 1 second at 100 samples/sec
    resamp = _resample(ac, epochs.info["sfreq"])
    resamp = resamp[:, 1:, np.newaxis, np.newaxis].transpose([2, 1, 3, 0])
    return np.real(resamp).astype(np.float32)


def _resample(ac: NDArray[float], fs: Union[int, float]) -> NDArray[float]:
    """Resample the autocorrelation feature.

    The comment in EEGLAB is:
        resample to 1 second at 100 samples/sec

    Which translates by: the output array must be of shape (n_comp, 101), thus
    the resampling up variable is set to 100, and down variable must respect:
        100 < ac.T.shape[0] * 100 / down <= 101
    If the instance sampling frequency is an integer, then down is equal to the
    sampling frequency.

    Parameters
    ----------
    ac : array
        Array of shape (n_comp, samples).
    fs : int | float
        Sampling frequency of the MNE instance.
    """
    down = int(fs)
    if 101 < ac.shape[1] * 100 / down:
        down += 1
    return resample_poly(ac.T, 100, down).T



import numpy as np
import pytest
from mne import create_info, make_fixed_length_epochs, pick_types
from mne.datasets import testing
from mne.io import RawArray, read_raw
from mne.preprocessing import ICA

from tmseegpy.mne_icalabel.config import ICA_LABELS_TO_MNE
from tmseegpy.mne_icalabel.iclabel import iclabel_label_components
from tmseegpy.mne_icalabel.utils._tests import requires_module

directory = testing.data_path() / "MEG" / "sample"
raw = read_raw(directory / "sample_audvis_trunc_raw.fif", preload=False)
raw.pick_types(eeg=True, exclude=[])
raw.load_data()
# preprocess, fake filtering, testing dataset is filtered between [0.1, 80] Hz
with raw.info._unlock():
    raw.info["highpass"] = 1.0
    raw.info["lowpass"] = 100.0
raw.set_eeg_reference("average")


@pytest.mark.parametrize(
    ("inst", "exclude"),
    [
        (raw, "bads"),
        (raw.copy().crop(0, 8), "bads"),
        (raw.copy().crop(0, 1), "bads"),
        (make_fixed_length_epochs(raw, duration=0.5, preload=True), "bads"),
        (make_fixed_length_epochs(raw, duration=1, preload=True), "bads"),
        (make_fixed_length_epochs(raw, duration=5, preload=True), "bads"),
        (raw, []),
        (raw.copy().crop(0, 8), []),
        (raw.copy().crop(0, 1), []),
        (make_fixed_length_epochs(raw, duration=0.5, preload=True), []),
        (make_fixed_length_epochs(raw, duration=1, preload=True), []),
        (make_fixed_length_epochs(raw, duration=5, preload=True), []),
    ],
)
@requires_module("onnxruntime")
def test_label_components_onnx(inst, exclude):
    """Check that label_components does not raise on various data shapes."""
    picks = pick_types(inst.info, eeg=True, exclude=exclude)
    ica = ICA(
        n_components=5, method="picard", fit_params=dict(ortho=False, extended=True)
    )
    ica.fit(inst, picks=picks)
    labels = iclabel_label_components(inst, ica, inplace=False, backend="onnx")
    assert labels.shape == (ica.n_components_, 7)
    assert len(ica.labels_) == 0
    labels2 = iclabel_label_components(inst, ica, inplace=True, backend="onnx")
    assert sorted(ica.labels_.keys()) == sorted(ICA_LABELS_TO_MNE.values())
    assert np.allclose(labels, labels2)


@pytest.mark.parametrize(
    ("inst", "exclude"),
    [
        (raw, "bads"),
        (raw.copy().crop(0, 8), "bads"),
        (raw.copy().crop(0, 1), "bads"),
        (make_fixed_length_epochs(raw, duration=0.5, preload=True), "bads"),
        (make_fixed_length_epochs(raw, duration=1, preload=True), "bads"),
        (make_fixed_length_epochs(raw, duration=5, preload=True), "bads"),
        (raw, []),
        (raw.copy().crop(0, 8), []),
        (raw.copy().crop(0, 1), []),
        (make_fixed_length_epochs(raw, duration=0.5, preload=True), []),
        (make_fixed_length_epochs(raw, duration=1, preload=True), []),
        (make_fixed_length_epochs(raw, duration=5, preload=True), []),
    ],
)
@requires_module("torch")
def test_label_components_torch(inst, exclude):
    """Check that label_components does not raise on various data shapes."""
    picks = pick_types(inst.info, eeg=True, exclude=exclude)
    ica = ICA(
        n_components=5, method="picard", fit_params=dict(ortho=False, extended=True)
    )
    ica.fit(inst, picks=picks)
    labels = iclabel_label_components(inst, ica, inplace=False, backend=None)
    assert labels.shape == (ica.n_components_, 7)
    assert len(ica.labels_) == 0
    labels2 = iclabel_label_components(inst, ica, inplace=True, backend=None)
    assert sorted(ica.labels_.keys()) == sorted(ICA_LABELS_TO_MNE.values())
    assert np.allclose(labels, labels2)

    picks = pick_types(inst.info, eeg=True, exclude=exclude)
    ica = ICA(
        n_components=5, method="picard", fit_params=dict(ortho=False, extended=True)
    )
    ica.fit(inst, picks=picks)
    labels = iclabel_label_components(inst, ica, inplace=False, backend="torch")
    assert labels.shape == (ica.n_components_, 7)
    assert len(ica.labels_) == 0
    labels2 = iclabel_label_components(inst, ica, inplace=True, backend="torch")
    assert sorted(ica.labels_.keys()) == sorted(ICA_LABELS_TO_MNE.values())
    assert np.allclose(labels, labels2)


def test_warnings(rng):
    """Test warnings issued when the instance does not meet the requirements."""
    times = np.linspace(0, 5, 2000)
    signals = np.array([np.sin(2 * np.pi * k * times) for k in (7, 22, 37)])
    coeffs = rng.random((6, 3))
    data = np.dot(coeffs, signals) + rng.normal(0, 0.1, (coeffs.shape[0], times.size))

    raw = RawArray(
        data,
        create_info(["Fpz", "Cz", "CPz", "Oz", "M1", "M2"], sfreq=400, ch_types="eeg"),
    )
    raw.set_montage("standard_1020")
    with raw.info._unlock():
        raw.info["highpass"] = 1.0

    # wrong raw, correct ica
    ica = ICA(
        n_components=3,
        method="infomax",
        fit_params=dict(extended=True),
        random_state=101,
    )
    ica.fit(raw)
    with (
        pytest.warns(RuntimeWarning, match="common average reference"),
        pytest.warns(RuntimeWarning, match="not filtered between 1 and 100 Hz"),
    ):
        iclabel_label_components(raw, ica)
    with raw.info._unlock():
        raw.info["lowpass"] = 100.0
    raw.set_eeg_reference("average")

    # infomax
    ica = ICA(
        n_components=3,
        method="infomax",
        fit_params=dict(extended=False),
        random_state=101,
    )
    ica.fit(raw)
    with pytest.warns(RuntimeWarning, match="designed with extended infomax ICA"):
        iclabel_label_components(raw, ica)
    # fastica
    ica = ICA(n_components=3, fit_params=dict(tol=1e-2), random_state=101)
    ica.fit(raw)
    with pytest.warns(RuntimeWarning, match="designed with extended infomax ICA"):
        iclabel_label_components(raw, ica)
    # picard
    ica = ICA(n_components=3, method="picard", random_state=101)
    ica.fit(raw)
    with pytest.warns(RuntimeWarning, match="designed with extended infomax ICA"):
        iclabel_label_components(raw, ica)

    raw = read_raw(directory / "sample_audvis_trunc_raw.fif", preload=False)
    raw.pick_types(meg=True)
    raw.load_data()
    ica = ICA(
        n_components=3,
        method="infomax",
        fit_params=dict(extended=True),
        random_state=101,
    )
    ica.fit(raw)
    # raw without EEG channels
    with pytest.raises(RuntimeError, match="Could not find EEG channels"):
        iclabel_label_components(raw, ica)

    epochs = make_fixed_length_epochs(raw)
    ica = ICA(
        n_components=3,
        method="infomax",
        fit_params=dict(extended=True),
        random_state=101,
    )
    ica.fit(epochs)
    # epochs without EEG channels
    with pytest.raises(RuntimeError, match="Could not find EEG channels"):
        iclabel_label_components(epochs, ica)


def test_comp_in_labels_():
    """Test that components already in labels_ are not added again."""
    picks = pick_types(raw.info, eeg=True, exclude="bads")
    ica = ICA(
        n_components=5,
        method="picard",
        fit_params=dict(ortho=False, extended=True),
        random_state=101,
    )
    ica.fit(raw, picks=picks)
    ica.labels_["brain"] = [3]
    ica.labels_["other"] = []
    iclabel_label_components(raw, ica, inplace=True)
    assert ica.labels_["brain"] == [3]
    assert ica.labels_["other"] == [1, 2, 4]



from __future__ import annotations  # c.f. PEP 563, PEP 649

import warnings
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from mne.io import BaseRaw
    from numpy.typing import ArrayLike, NDArray


def _mne_to_eeglab_locs(
    raw: BaseRaw, picks: list[str]
) -> tuple[NDArray[float], NDArray[float]]:
    """Obtain EEGLab-like spherical coordinate from EEG channel positions.

    TODO: @JACOB:
    - Where is (0,0,0) defined in MNE vs EEGLab?
    - some text description of how the sphere coordinates differ between MNE
    and EEGLab.

    Parameters
    ----------
    raw : mne.io.BaseRaw
        Instance of raw object with a `mne.montage.DigMontage` set with
        ``n_channels`` channel positions.
    picks : list of str
        List of channel names to include.

    Returns
    -------
    Rd : np.array of shape (1, n_channels)
        Angle in spherical coordinates of each EEG channel.
    Th : np.array of shape (1, n_channels)
        Degree in spherical coordinates of each EEG channel.
    """

    def _sph2topo(_theta, _phi):
        """Convert spherical coordinates to topo."""
        az = _phi
        horiz = _theta
        angle = -1 * horiz
        radius = (np.pi / 2 - az) / np.pi
        return angle, radius

    def _cart2sph(_x, _y, _z):
        """Convert cartesian coordinates to spherical."""
        azimuth = np.arctan2(_y, _x)
        elevation = np.arctan2(_z, np.sqrt(_x**2 + _y**2))
        r = np.sqrt(_x**2 + _y**2 + _z**2)
        # theta,phi,r
        return azimuth, elevation, r

    # get the channel position dictionary
    montage = raw.copy().pick(picks).get_montage()
    if montage is None:
        raise ValueError(
            "Montage is not set. Please set the montage to provide the electrode "
            "positions."
        )
    positions = montage.get_positions()
    ch_pos = positions["ch_pos"]
    # check that we do have a coordinate for every points
    empty = [key for key in ch_pos if np.all(np.isnan(ch_pos[key]))]
    if len(empty) != 0:
        raise ValueError(
            f"Channel position for {empty} is missing. Please check the montage set, "
            "the channel names, and ensure that every channel has a position."
        )

    # get locations as a 2D array
    locs = np.vstack(list(ch_pos.values()))

    # Obtain cartesian coordinates
    x = locs[:, 1]

    # be mindful of the nose orientation in eeglab and mne
    # TODO: @Jacob, please expand on this.
    y = -1 * locs[:, 0]
    # see https://github.com/mne-tools/mne-python/blob/24377ad3200b6099ed47576e9cf8b27578d571ef/mne/io/eeglab/eeglab.py#L105  # noqa
    z = locs[:, 2]

    # Obtain Spherical Coordinates
    sph = np.array([_cart2sph(x[i], y[i], z[i]) for i in range(len(x))])
    theta = sph[:, 0]
    phi = sph[:, 1]

    # Obtain Polar coordinates (as in eeglab)
    topo = np.array([_sph2topo(theta[i], phi[i]) for i in range(len(theta))])
    rd = topo[:, 1]
    th = topo[:, 0]

    return rd.reshape([1, -1]), np.degrees(th).reshape([1, -1])


def _pol2cart(
    theta: NDArray[float], rho: NDArray[float]
) -> tuple[NDArray[float], NDArray[float]]:
    """Convert polar coordinates to cartesian coordinates.

    Parameters
    ----------
    theta : array
        angle
    rho : array
        magnitude
    """
    x = rho * np.cos(theta)
    y = rho * np.sin(theta)
    return x, y


# ----------------------------------------------------------------------------
def _next_power_of_2(x) -> int:
    """Equivalent to 2^nextpow2 in MATLAB."""
    return 1 if x == 0 else 2 ** (x - 1).bit_length()


# ----------------------------------------------------------------------------
def _gdatav4(
    x: ArrayLike, y: ArrayLike, v: ArrayLike, xq: ArrayLike, yq: ArrayLike
) -> tuple[ArrayLike, ArrayLike, ArrayLike]:
    """GDATAV4 MATLAB 4 GRIDDATA interpolation.

    Parameters
    ----------
    x : array
        x-coordinates
    y : array
        y-coordinates
    v : array
        values
    xq : array
        x-grid
    yq : array
        y-grid

    Returns
    -------
    xq : array
    yq : array
    vq : array

    Reference
    ---------
    David T. Sandwell, Biharmonic spline interpolation of GEOS-3 and SEASAT
    altimeter data, Geophysical Research Letters, 2, 139-142, 1987.

    Describes interpolation using value of gradient of value in any dimension.
    """
    x, y, v = _mergepoints2D(x, y, v)

    xy = x + 1j * y
    xy = np.squeeze(xy)

    # Determine distances between points
    d = np.abs(np.subtract.outer(xy, xy))
    # Determine weights for interpolation
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message="divide by zero encountered in log",
            category=RuntimeWarning,
        )
        warnings.filterwarnings(
            "ignore",
            message="invalid value encountered in multiply",
            category=RuntimeWarning,
        )
        g = np.square(d) * (np.log(d) - 1)  # Green's function.
    # Fixup value of Green's function along diagonal
    np.fill_diagonal(g, 0)
    weights = np.linalg.lstsq(g, v, rcond=-1)[0]

    m, n = xq.shape
    vq = np.zeros(xq.shape)

    # Evaluate at requested points (xq,yq). Loop to save memory.
    for i in range(m):
        for j in range(n):
            d = np.abs(xq[i, j] + 1j * yq[i, j] - xy)
            g = np.square(d) * (np.log(d) - 1)
            # Value of Green's function at zero
            g[np.where(np.isclose(d, 0))] = 0
            vq[i, j] = (np.expand_dims(g, axis=0) @ np.expand_dims(weights, axis=1))[0][
                0
            ]
    return xq, yq, vq


def _mergepoints2D(
    x: ArrayLike, y: ArrayLike, v: ArrayLike
) -> tuple[ArrayLike, ArrayLike, ArrayLike]:
    """Averages values for points that are close to each other.

    Parameters
    ----------
    x : array
        x-coordinates
    y : array
        y-coordinates
    v : array
        values

    Returns
    -------
    x : array
    y : array
    v : array
    """
    # Sort x and y so duplicate points can be averaged
    # Need x,y and z to be column vectors
    sz = x.size
    x = x.copy()
    y = y.copy()
    v = v.copy()
    x = np.reshape(x, sz, order="F")
    y = np.reshape(y, sz, order="F")
    v = np.reshape(v, sz, order="F")

    myepsx = np.spacing(0.5 * (np.max(x) - np.min(x))) ** (1 / 3)
    myepsy = np.spacing(0.5 * (np.max(y) - np.min(y))) ** (1 / 3)
    # Look for x, y points that are identical (within a tolerance)
    # Average out the values for these points
    if np.all(np.isreal(v)):
        data = np.stack((y, x, v), axis=-1)
        yxv = _mergesimpts(data, [myepsy, myepsx, np.inf], "average")
        x = yxv[:, 1]
        y = yxv[:, 0]
        v = yxv[:, 2]
    else:
        # If z is imaginary split out the real and imaginary parts
        data = np.stack((y, x, np.real(v), np.imag(v)), axis=-1)
        yxv = _mergesimpts(data, [myepsy, myepsx, np.inf, np.inf], "average")
        x = yxv[:, 1]
        y = yxv[:, 0]
        # Re-combine the real and imaginary parts
        v = yxv[:, 2] + 1j * yxv[:, 3]

    return x, y, v


def _mergesimpts(
    data: ArrayLike, tols: list[ArrayLike], mode: str = "average"
) -> ArrayLike:
    """Merge similar points."""
    data_ = data.copy()[np.argsort(data[:, 0])]
    newdata = []
    tols_ = np.array(tols)
    idxs_ready: list[int] = []
    point = 0
    for point in range(data_.shape[0]):
        if point in idxs_ready:
            continue
        else:
            similar_pts = np.where(
                np.prod(np.abs(data_ - data_[point]) < tols_, axis=-1)
            )
            similar_pts = np.array(list(set(similar_pts[0].tolist()) - set(idxs_ready)))
            idxs_ready += similar_pts.tolist()
            if mode == "average":
                exemplar = np.mean(data_[similar_pts], axis=0)
            else:
                exemplar = data_[similar_pts].copy()[0]  # first
            newdata.append(exemplar)
    return np.array(newdata)



import numpy as np
import pytest
from mne import set_log_level


def pytest_configure(config):
    """Configure pytest options."""
    warnings_lines = r"""
    error::
    # Until MNE 1.6 is the minimum supported version
    # c.f. https://github.com/mne-tools/mne-python/pull/12143
    ignore:Setting non-standard config type.*:RuntimeWarning
    # pandas 3.0 will require pyarrow
    ignore:\n*Pyarrow will become a required dependency of pandas.*:DeprecationWarning
    # Python 3.12+ gives a deprecation warning if TarFile.extraction_filter is None
    ignore:Python 3\.14 will, by default, filter extracted tar.*:DeprecationWarning
    # onnxruntime on windows runners
    ignore:Unsupported Windows version.*:UserWarning
    """
    for warning_line in warnings_lines.split("\n"):
        warning_line = warning_line.strip()
        if warning_line and not warning_line.startswith("#"):
            config.addinivalue_line("filterwarnings", warning_line)
    set_log_level("WARNING")


@pytest.fixture(scope="session")
def rng():
    """Return a numpy random generator."""
    return np.random.default_rng()



import argparse

from mne import read_epochs
from mne.io import read_raw
from mne.preprocessing import read_ica
from qtpy.QtWidgets import QApplication

from mne_icalabel.gui._label_components import ICAComponentLabeler


def main():
    """Entry point for mne_gui_ic_annotation."""
    parser = argparse.ArgumentParser(
        prog="mne-icalabel", description="IC annotation GUI"
    )
    parser.add_argument(
        "fname_raw_epo",
        type=str,
        help="path to the file with raw or epochs.",
    )
    parser.add_argument(
        "fname_ica",
        type=str,
        help="path to the file with the ICA decomposition.",
    )
    parser.add_argument("--dev", help="loads a sample dataset.", action="store_true")
    args = parser.parse_args()

    if not args.dev:
        for func in (read_raw, read_epochs):
            try:
                inst = func(args.fname_raw_epo, preload=True)
                break
            except Exception:
                pass
        else:
            raise RuntimeError(f"Could not load the file {args.fname_raw_epo}.")
        ica = read_ica(args.fname_ica)
    else:
        from mne.datasets import sample
        from mne.preprocessing import ICA

        directory = sample.data_path() / "MEG" / "sample"
        inst = read_raw(directory / "sample_audvis_raw.fif", preload=False)
        inst.crop(0, 10).pick_types(eeg=True, exclude="bads")
        inst.load_data()
        # preprocess
        inst.filter(l_freq=1.0, h_freq=100.0)
        inst.set_eeg_reference("average")

        n_components = 15
        ica = ICA(n_components=n_components, method="picard")
        ica.fit(inst)

    app = QApplication([])
    window = ICAComponentLabeler(inst=inst, ica=ica)
    window.show()
    app.exec()



from __future__ import annotations  # c.f. PEP 563, PEP 649

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from numpy.typing import ArrayLike, NDArray


def _format_input(
    topo: ArrayLike, psd: ArrayLike, autocorr: ArrayLike
) -> tuple[NDArray, NDArray, NDArray]:
    """Replicate the input formatting in EEGLAB -ICLabel.

    .. code-block:: matlab

       images = cat(4, images, -images, images(:, end:-1:1, :, :), ...
                    -images(:, end:-1:1, :, :));
       psds = repmat(psds, [1 1 1 4]);
       autocorrs = repmat(autocorrs, [1 1 1 4]);
    """
    formatted_topo = np.concatenate(
        (topo, -1 * topo, np.flip(topo, axis=1), np.flip(-1 * topo, axis=1)),
        axis=3,
    )
    formatted_psd = np.tile(psd, (1, 1, 1, 4))
    formatted_autocorr = np.tile(autocorr, (1, 1, 1, 4))
    return formatted_topo, formatted_psd, formatted_autocorr



import numpy as np
from numpy import linalg
import scipy.signal
from typing import Optional, Union, Dict, List, Tuple
import matplotlib.pyplot as plt
import mne
from matplotlib.gridspec import GridSpec
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable
from mpl_toolkits.mplot3d import proj3d


class PCIst:
    """
    Class implementation of PCIst (Perturbational Complexity Index based on State transitions).
    
    This implementation is a copy paste from:
    Comolatti R et al., "A fast and general method to empirically estimate the complexity of brain responses
    to transcranial and intracranial stimulations" Brain Stimulation (2019):
    https://github.com/renzocom/PCIst/tree/master
    
    Parameters
    ----------
    epochs : mne.Epochs
        MNE epochs object containing TMS-evoked potentials. 
        Should be preprocessed before passing to this class.
    """
    
    def __init__(self, epochs: mne.epochs.Epochs):
        if not isinstance(epochs, mne.epochs.Epochs):
            print(f"Type of epochs: {type(epochs)}")
            raise ValueError("Input must be an MNE Epochs object")
        self.epochs = epochs
        self.PCI = None

    def calc_PCIst(self, full_return=False, return_details=True, **par):
        ''' Calculates PCIst (Perturbational Complexity Index based on State transitions) of a signal.
        Parameters
        ----------
        epochs : mne.epochs object
            Preprocessed mne.epochs object.
        times : ndarray
            1D array (time,) containing timepoints (negative values are baseline).
        full_return : bool
            Returns multiple variables involved in PCI computation.
        **pars : dictionary
            Dictionary containing parameters (see dimensionality_reduction(),
            state_transition_quantification()and preprocess_signal() documentation).
            Example:
            >> par = {'baseline_window':(-400,-50), 'response_window':(0,300), 'k':1.2, 'min_snr':1.1,
            'max_var':99, 'embed':False,'n_steps':100}
            >> PCIst, PCIst_bydim = calc_PCIst(signal_evoked, times, **par)

        Returns
        -------
        float
            PCIst value
        OR (if full_return==True)
        dict
            Dictionary containing all variables from calculation including array 'dNSTn' with PCIst decomposition.
        '''
        # Get evoked data
        self.evoked = self.epochs.average()    
        # Get data and times and reshape
        self.signal_evk = self.evoked.get_data()
        self.times = self.evoked.times * 1000
        
        t_data = self.evoked.data
        print(f"Input data shape: {t_data.shape}")  # Should be (channels, timepoints)
        print(f"Number of channels: {len(self.epochs.ch_names)}")
        print(f"Number of timepoints: {len(self.evoked.times)}")

        if np.any(np.isnan(self.signal_evk)):
            print('Data contains nan values.')
            return 0

        #signal_evk, times = preprocess_signal(signal_evk, times, (par['baseline_window'][0],
        #                                                        par['response_window'][1]), **par)
        
        self.baseline_window = par.get('baseline_window')
        self.response_window = par.get('response_window')
        
        self.signal_svd, self.var_exp, self.eigenvalues, self.snrs = self.dimensionality_reduction(
            signal=self.signal_evk,
            times=self.times,
            response_window=self.response_window,
            baseline_window=self.baseline_window,
            max_var=par.get('max_var', 99),
            min_snr=par.get('min_snr', 1.1),
            n_components=par.get('n_components', None)
        )
        print(f"Number of components retained: {self.signal_svd.shape[0]}")
        print(f"Variance explained: {np.sum(self.var_exp):.2f}%")
        print(f"SNRs of components: {self.snrs}")

        STQ = self.state_transition_quantification(
            signal=self.signal_svd,
            times=self.times,
            k=par.get('k', 1.2),
            baseline_window=self.baseline_window,
            response_window=self.response_window,
            embed=par.get('embed', False),
            L=par.get('L', None),
            tau=par.get('tau', None),
            n_steps=par.get('n_steps', 100),
            max_thr_p=par.get('max_thr_p', 1.0)
        )
        print(f"NST base range: {np.min(STQ['NST_base']):.3f} to {np.max(STQ['NST_base']):.3f}")
        print(f"NST resp range: {np.min(STQ['NST_resp']):.3f} to {np.max(STQ['NST_resp']):.3f}")
        print(f"dNST values: {STQ['dNST']}")

        PCI = np.sum(STQ['dNST'])
        self.PCI = PCI 

        if full_return:
            return {'PCI':PCI, **STQ, 'signal_evk':self.signal_evk, 'times':self.times, 'signal_svd':self.signal_svd,
                    'eigenvalues':self.eigenvalues, 'var_exp':self.var_exp, 'snrs':self.snrs}
        
        if return_details:
            details = {
                'components': self.signal_svd,
                'times': self.times,
                'evoked_data': self.signal_evk,
                'eigenvalues': self.eigenvalues,
                'var_exp': self.var_exp,
                'snr_values': self.snrs,
                'distance_matrices': [(STQ['D_base'][0], STQ['D_resp'][0])],
                'transition_matrices': [(STQ['T_base'][0], STQ['T_resp'][0])],
                'optimal_thresholds': [STQ['max_thresholds'][0]],
                'thresholds': [STQ['thresholds'][:,0]],
                'nst_base_values': [STQ['NST_base'][:,0]],
                'nst_resp_values': [STQ['NST_resp'][:,0]],
                'diff_values': [STQ['NST_diff'][:,0]],
                'delta_nst_values': [STQ['dNST'][0]],
                'baseline_window': self.baseline_window,
                'response_window': self.response_window 
            }
            return PCI, details
            
        return PCI

    ## DIMENSIONALITY REDUCTION
    def dimensionality_reduction(self, signal, times, response_window, baseline_window, max_var=99, min_snr=1.1, n_components=None):
        print(f"Dimensionality Reduction Parameters:")
        print(f"  response_window: {response_window}")
        print(f"  baseline_window: {baseline_window}")
        print(f"  max_var: {max_var}")
        print(f"  min_snr: {min_snr}")
        print(f"  n_components: {n_components}")
        '''Returns principal components of signal according to SVD of the response.

        Calculates SVD at a given time interval (response_window) and uses the new basis to transform
        the whole signal yielding `n_components` principal components. The principal components are
        then selected to account for at least `max_var`% of the variance basesent in the signal's
        response.

        Parameters
        ----------
        signal : ndarray
            2D array (ch,time) containing signal.
        times : ndarray
            1D array (time,) containing timepoints
        response_window : tuple
            Signal's response time interval (ini,end).
        max_var: 0 < float <= 100
            Percentage of variance accounted for by the selected principal components.
        min_snr : float, optional
            Selects principal components with a signal-to-noise ratio (SNR) > min_snr.
        n_components : int, optional
            Number of principal components calculated (before selection).


        Returns
        -------
        np.ndarray
            2D array (ch,time) with selected principal components.
        np.ndarray
            1D array (n_components,) with `n_components` SVD eigenvalues of the signal's response.
        '''
        print(f"Dimensionality Reduction Parameters:")
        print(f"  response_window: {response_window}")
        print(f"  baseline_window: {baseline_window}")
        print(f"  max_var: {max_var}")
        print(f"  min_snr: {min_snr}")
        print(f"  n_components: {n_components}")

        if not n_components:
            n_components = signal.shape[0]

        Vk, eigenvalues = self.get_svd(signal, times, response_window, n_components)
        var_exp = 100 * eigenvalues**2/np.sum(eigenvalues**2)

        signal_svd = self.apply_svd(signal, Vk)

        max_dim = self.calc_maxdim(eigenvalues, max_var)

        signal_svd = signal_svd[:max_dim, :]

        snrs = self.calc_snr(signal_svd, times, baseline_window, response_window)
        signal_svd = signal_svd[snrs > min_snr, :]
        snrs = snrs[snrs > min_snr]

        Nc = signal_svd.shape[0]

        # if min_snr:
            # base_ini_ix = get_time_index(times, kwargs['baseline_window'][0])
            # base_end_ix = get_time_index(times, kwargs['baseline_window'][1])
            # resp_ini_ix = get_time_index(times, response_window[0])
            # resp_end_ix = get_time_index(times, response_window[1])
            # n_dims = np.size(signal_svd, 0)
            # snrs = np.zeros(n_dims)
            # for c in range(n_dims):
            #     resp_power = np.mean(np.square(signal_svd[c, resp_ini_ix:resp_end_ix]))
            #     base_power = np.mean(np.square(signal_svd[c, base_ini_ix:base_end_ix]))
            #     snrs[c] = np.sqrt(np.divide(resp_power, base_power))

        return signal_svd, var_exp[:Nc], eigenvalues, snrs

    def calc_snr(self, signal_svd, times, baseline_window, response_window):

        base_ini_ix = self.get_time_index(times, baseline_window[0])
        base_end_ix = self.get_time_index(times, baseline_window[1])
        resp_ini_ix = self.get_time_index(times, response_window[0])
        resp_end_ix = self.get_time_index(times, response_window[1])

        resp_power = np.mean(np.square(signal_svd[:,resp_ini_ix:resp_end_ix]), axis=1)
        base_power = np.mean(np.square(signal_svd[:,base_ini_ix:base_end_ix]), axis=1)
        snrs = np.sqrt(resp_power / base_power)
        return snrs
    

    def get_svd(self, signal_evk, times, response_window, n_components):
        ini_t, end_t = response_window
        ini_ix = self.get_time_index(times, onset=ini_t)
        end_ix = self.get_time_index(times, onset=end_t)
        signal_resp = signal_evk[:, ini_ix:end_ix].T
        print(f"Signal shape before SVD: {signal_resp.shape}")
        U, S, V = linalg.svd(signal_resp, full_matrices=False)
        print(f"All singular values: {S}")
        print(f"Number of non-zero singular values: {np.sum(S > 1e-10)}")
        V = V.T
        Vk = V[:, :n_components]
        eigenvalues = S[:n_components]
        return Vk, eigenvalues

    def apply_svd(self, signal, V):
        '''Transforms signal according to SVD basis.'''
        return signal.T.dot(V).T

    ## STATE TRANSITION QUANTIFICATION
    def state_transition_quantification(self, signal, times, k, baseline_window, response_window, embed=False,
                                        L=None, tau=None, n_steps=100, max_thr_p=1.0, **kwargs):
        ''' Receives selected principal components of perturbational signal and
        performs state transition quantification.

        Parameters
        ----------
        signal : ndarray
            2D array (component,time) containing signal (typically, the selected
            principal components).
        times : ndarray
            1D array (time,) containing timepoints
        k : float > 1
            Noise control parameter.
        baseline_window : tuple
            Signal's baseline time interval (ini,end).
        response_window : tuple
            Signal's response time interval (ini,end).
        embed : bool, optional
            Perform time-delay embedding.
        L : int
            Number of embedding dimensions.
        tau : int
            Number of timesamples of embedding delay
        n_steps : int, optional
            Number of steps used to search for the threshold that maximizes ∆NST.
            Search is performed by partitioning  the interval (defined from the median
            of the baseline’s distance matrix to the maximum of the response’s
            distance matrix) into ‘n_steps’ equal lengths.

        Returns
        -------
        float
            PCIst value.
        ndarray
            List containing component wise PCIst value (∆NSTn).
        '''
        print(f"State Transition Quantification Parameters:")
        print(f"  k: {k}")
        print(f"  baseline_window: {baseline_window}")
        print(f"  response_window: {response_window}")
        print(f"  embed: {embed}")
        print(f"  L: {L}")
        print(f"  tau: {tau}")
        print(f"  n_steps: {n_steps}")
        print(f"  max_thr_p: {max_thr_p}")
        n_dims = signal.shape[0]
        if n_dims == 0:
            print('No components --> PCIst=0')
            return {'dNST':np.array([]), 'n_dims':0}

        # EMBEDDING
        if embed:
            cut = (L-1)*tau
            times = times[cut:]
            temp_signal = np.zeros((n_dims, L, len(times)))
            for i in range(n_dims):
                temp_signal[i, :, :] = self.dimension_embedding(signal[i, :], L, tau)
            signal = temp_signal

        else:
            signal = signal[:, np.newaxis, :]

        # BASELINE AND RESPONSE DEFINITION
        base_ini_ix = self.get_time_index(times, baseline_window[0])
        base_end_ix = self.get_time_index(times, baseline_window[1])
        resp_ini_ix = self.get_time_index(times, response_window[0])
        resp_end_ix = self.get_time_index(times, response_window[1])
        n_baseline = len(times[base_ini_ix:base_end_ix])
        n_response = len(times[resp_ini_ix:resp_end_ix])

        if n_response <= 1 or n_baseline <= 1:
            print('Warning: Bad time interval defined.')

        baseline = signal[:, :, base_ini_ix:base_end_ix]
        response = signal[:, :, resp_ini_ix:resp_end_ix]

        # NST CALCULATION
            # Distance matrix
        D_base = np.zeros((n_dims, n_baseline, n_baseline))
        D_resp = np.zeros((n_dims, n_response, n_response))
            # Transition matrix
        T_base = np.zeros((n_steps, n_dims, n_baseline, n_baseline))
        T_resp = np.zeros((n_steps, n_dims, n_response, n_response))
            # Number of state transitions
        NST_base = np.zeros((n_steps, n_dims))
        NST_resp = np.zeros((n_steps, n_dims))
        thresholds = np.zeros((n_steps, n_dims))
        for i in range(n_dims):
            D_base[i, :, :] = self.recurrence_matrix(baseline[i, :, :], thr=None, mode='distance')
            D_resp[i, :, :] = self.recurrence_matrix(response[i, :, :], thr=None, mode='distance')
            min_thr = np.median(D_base[i, :, :].flatten())
            max_thr = np.max(D_resp[i, :, :].flatten()) * max_thr_p
            thresholds[:, i] = np.linspace(min_thr, max_thr, n_steps)
        for i in range(n_steps):
            for j in range(n_dims):
                T_base[i, j, :, :] = self.distance2transition(D_base[j, :, :], thresholds[i, j])
                T_resp[i, j, :, :] = self.distance2transition(D_resp[j, :, :], thresholds[i, j])

                NST_base[i, j] = np.sum(T_base[i, j, :, :])/n_baseline**2
                NST_resp[i, j] = np.sum(T_resp[i, j, :, :])/n_response**2

        # PCIST
        NST_diff = NST_resp - k * NST_base
        ixs = np.argmax(NST_diff, axis=0)
        max_thresholds = np.array([thresholds[ix, i] for ix, i in zip(ixs, range(n_dims))])
        dNST = np.array([NST_diff[ix, i] for ix, i in zip(ixs, range(n_dims))]) * n_response
        dNST = [x if x>0 else 0 for x in dNST]

        temp = np.zeros((n_dims, n_response, n_response))
        temp2 = np.zeros((n_dims, n_baseline, n_baseline))
        for i in range(n_dims):
            temp[i, :, :] = T_resp[ixs[i], i, :, :]
            temp2[i, :, :] = T_base[ixs[i], i, :, :]
        T_resp = temp
        T_base = temp2

        return {'dNST':dNST, 'n_dims':n_dims,
        'D_base':D_base, 'D_resp':D_resp, 'T_base':T_base,'T_resp':T_resp,
        'thresholds':thresholds, 'NST_diff':NST_diff, 'NST_resp':NST_resp, 'NST_base':NST_base,'max_thresholds':max_thresholds}


    def recurrence_matrix(self, signal, mode, thr=None):
        ''' Calculates distance, recurrence or transition matrix. Signal can be
        embedded (m, n_times) or not (, n_times).

        Parameters
        ----------
        signal : ndarray
            Time-series; may be a 1D (time,) or a m-dimensional array (m, time) for
            time-delay embeddeding.
        mode : str
            Specifies calculated matrix: 'distance', 'recurrence' or 'transition'
        thr : float, optional
            If transition matrix is chosen (`mode`=='transition'), specifies threshold value.

        Returns
        -------
        ndarray
            2D array containing specified matrix.
        '''
        if len(signal.shape) == 1:
            signal = signal[np.newaxis, :]
        n_dims = signal.shape[0]
        n_times = signal.shape[1]

        R = np.zeros((n_dims, n_times, n_times))
        for i in range(n_dims):
            D = np.tile(signal[i, :], (n_times, 1))
            D = D - D.T
            R[i, :, :] = D
        R = np.linalg.norm(R, ord=2, axis=0)

        mask = (R <= thr) if thr else np.zeros(R.shape).astype(bool)
        if mode == 'distance':
            R[mask] = 0
            return R
        if mode == 'recurrence':
            return mask.astype(int)
        if mode == 'transition':
            return self.diff_matrix(mask.astype(int), symmetric=False)
        return 0

    def distance2transition(self, dist_R, thr):
        ''' Receives 2D distance matrix and calculates transition matrix. '''
        mask = dist_R <= thr
        R = self.diff_matrix(mask.astype(int), symmetric=False)
        return R

    def distance2recurrence(self, dist_R, thr):
        ''' Receives 2D distance matrix and calculates recurrence matrix. '''
        mask = dist_R <= thr
        return mask.astype(int)

    def diff_matrix(self, A, symmetric=False):
        B = np.abs(np.diff(A))
        if B.shape[1] != B.shape[0]:
            B2 = np.zeros((B.shape[0], B.shape[1]+1))
            B2[:, :-1] = B
            B = B2
        if symmetric:
            B = (B + B.T)
            B[B > 0] = 1
        return B

    def calc_maxdim(self, eigenvalues, max_var):
        ''' Get number of dimensions that accumulates at least `max_var`% of total variance'''
        if max_var == 100:
            return len(eigenvalues)
        eigenvalues = np.sort(eigenvalues)[::-1] # Sort in descending order
        var = eigenvalues ** 2
        var_p = 100 * var/np.sum(var)
        var_cum = np.cumsum(var_p)
        max_dim = len(eigenvalues) - np.sum(var_cum >= max_var) + 1
        return max_dim

    def dimension_embedding(self, x, L, tau):
        '''
        Returns time-delay embedding of vector.
        Parameters
        ----------
        x : ndarray
            1D array time series.
        L : int
            Number of dimensions in the embedding.
        tau : int
            Number of samples in delay.
        Returns
        -------
        ndarray
            2D array containing embedded signal (L, time)

        '''
        assert len(x.shape) == 1, "x must be one-dimensional array (n_times,)"
        n_times = x.shape[0]
        s = np.zeros((L, n_times - (L-1) * tau))
        ini = (L-1) * tau if L > 1 else None
        s[0, :] = x[ini:]
        for i in range(1, L):
            ini = (L-i-1) * tau
            end = -i * tau
            s[i, :] = x[ini:end]
        return s

    ## PREPROCESS
    def preprocess_signal(self, signal_evk, times, time_window, baseline_corr=False, resample=None,
                        avgref=False, **kwargs):
        assert signal_evk.shape[1] == len(times), 'Signal and Time arrays must be of the same size.'
        if avgref:
            signal_evk = self.avgreference(signal_evk)
        if baseline_corr:
            signal_evk = self.baseline_correct(signal_evk, times, delta=-50)
        t_ini, t_end = time_window
        ini_ix = self.get_time_index(times, t_ini)
        end_ix = self.get_time_index(times, t_end)
        signal_evk = signal_evk[:, ini_ix:end_ix]
        times = times[ini_ix:end_ix]
        if resample:
            signal_evk, times = self.undersample_signal(signal_evk, times, new_fs=resample)
        return signal_evk, times

    def avgreference(self, signal):
        ''' Performs average reference to signal. '''
        new_signal = np.zeros(signal.shape)
        channels_mean = np.mean(signal, axis=0)[np.newaxis]
        new_signal = signal - channels_mean
        return new_signal

    def undersample_signal(self, signal, times, new_fs):
        '''
        signal : (ch x times)
        times : (times,) [ms]
        new_fs : [hz]
        '''
        n_samples = int((times[-1]-times[0])/1000 * new_fs)
        new_signal_evk, new_times = scipy.signal.resample(signal, n_samples, t=times, axis=1)
        return new_signal_evk, new_times

    def baseline_correct(self, Y, times, delta=0):
        ''' Baseline correct signal using times < delta '''
        newY = np.zeros(Y.shape)
        onset_ix = self.get_time_index(times, delta)
        baseline_mean = np.mean(Y[:, :onset_ix], axis=1)[np.newaxis]
        newY = Y - baseline_mean.T
        close_enough = np.all(np.isclose(np.mean(newY[:, :onset_ix], axis=1), 0, atol=1e-08))
        assert close_enough, "Baseline mean is not zero"
        return newY

    def get_time_index(self, times, onset=0):
        ''' Returns index of first time greater then delta. For delta=0 gets index of
        first non-negative time.
        '''
        return np.sum(times < onset)
    
    def plot_analysis(self, details: Dict, session_name: str = None) -> plt.Figure:
        """
        Plot PCIst analysis steps with parallel plots for baseline/response periods.
        Includes dual thresholds, enhanced projections, and visual connections.
        """
        # Input validation
        required_keys = ['components', 'times', 'evoked_data', 'eigenvalues', 'var_exp']
        if not all(key in details for key in required_keys):
            raise ValueError("Missing required keys in details dictionary")
        
        # Get number of available components
        n_available_components = details['components'].shape[0]
        # Create larger figure and adjust height ratios for bigger surfaces
        fig = plt.figure(figsize=(12, 20))
        gs = plt.GridSpec(8, 2, height_ratios=[0.5, 1.5, 0.75, 0.75, 1.5, 2.5, 2, 1])

        def add_connecting_lines(ax_matrix, ax_surface, fig):
            xlim = ax_matrix.get_xlim()
            ylim = ax_matrix.get_ylim()

            # Define corners in the correct order
            matrix_corners = [
                [xlim[0], ylim[0]],  # Lower-left corner
                [xlim[1], ylim[0]],  # Lower-right corner
                [xlim[1], ylim[1]],  # Upper-right corner
                [xlim[0], ylim[1]],  # Upper-left corner
            ]
            # Transform 2D data coordinates to display coordinates
            matrix_corners_display = [ax_matrix.transData.transform(c) for c in matrix_corners]

            # Project 3D corners to 2D display coordinates
            surface_corners_display = []
            for x3d, y3d in matrix_corners:
                z3d = ax_surface.get_zlim()[0]  # Use the bottom z-limit
                x2d, y2d, _ = proj3d.proj_transform(x3d, y3d, z3d, ax_surface.get_proj())
                x_disp, y_disp = ax_surface.transData.transform((x2d, y2d))
                surface_corners_display.append([x_disp, y_disp])

            # Draw lines between corresponding corners
            for (mx, my), (sx, sy) in zip(matrix_corners_display, surface_corners_display):
                mx_fig, my_fig = fig.transFigure.inverted().transform((mx, my))
                sx_fig, sy_fig = fig.transFigure.inverted().transform((sx, sy))
                line = plt.Line2D([mx_fig, sx_fig],
                                [my_fig, sy_fig],
                                transform=fig.transFigure, color='gray', linestyle='--', alpha=0.5)
                fig.lines.append(line)


        # Add text box with analysis details at the top
        ax_text = fig.add_subplot(gs[0, :])
        analysis_text = (
            f"Session name: {session_name}\n"
            f"Number of non-zero singular values: {np.sum(details['eigenvalues'] > 1e-10)}\n"
            f"Number of components retained: {details['components'].shape[0]}\n"
            f"Variance explained: {np.sum(details['var_exp']):.2f}%\n"
            f"PCI: {self.PCI:.6f}"
        )
        ax_text.text(0.05, 0.05, analysis_text, fontsize=10, family='monospace',
                    verticalalignment='bottom', transform=ax_text.transAxes)
        ax_text.axis('off')
        
        # A: TMS-evoked potentials
        ax1 = fig.add_subplot(gs[1, :])
        ax1.plot(details['times'], details['evoked_data'].T, 'b-', alpha=0.3, linewidth=0.5)
        ax1.set_title('A) TMS-Evoked Potentials')
        ax1.axvline(x=0, color='k', linestyle='--')
        ax1.text(0, ax1.get_ylim()[1], 'TMS', ha='center', va='bottom')
        ax1.text(-350, ax1.get_ylim()[1], 'TEPs', va='center', ha='right')
        
        max_pcs_to_plot = min(2, n_available_components)  # Plot up to 2 PCs if available
        for i in range(max_pcs_to_plot):
            ax_pc = fig.add_subplot(gs[2 + i, :])
            if i == 0:
                ax_pc.plot(details['times'], details['components'][i], 'k-')
                ax_pc.text(-350, ax_pc.get_ylim()[1], 'Principal\nComponents', 
                        va='center', ha='right')
            else:
                ax_pc.plot(details['times'], details['components'][i], 'b-')
            ax_pc.text(-300, ax_pc.get_ylim()[1], f'PC{i+1}', va='top')
            ax_pc.axvline(x=0, color='k', linestyle='--', alpha=0.5)
        
        # If we have fewer than 2 components, fill remaining subplot with informative text
        if n_available_components < 2:
            for i in range(n_available_components, 2):
                ax_empty = fig.add_subplot(gs[2 + i, :])
                ax_empty.text(0.5, 0.5, f'PC{i+1} not available\n(only {n_available_components} component{"s" if n_available_components != 1 else ""} present)',
                            ha='center', va='center')
                ax_empty.axis('off')
        
        # Get common scale for distance matrices and surfaces
        baseline_dist, response_dist = details['distance_matrices'][0]
        vmin = min(baseline_dist.min(), response_dist.min())
        vmax = max(baseline_dist.max(), response_dist.max())
        
        # Get thresholds
        optimal_threshold = details['optimal_thresholds'][0]
        lower_threshold = optimal_threshold * 0.6
        
        # B: Distance matrices
        # Baseline distance matrix
        # Baseline distance matrix
        ax3 = fig.add_subplot(gs[4, 0])
        im1 = ax3.imshow(baseline_dist, aspect='equal', cmap='viridis',
                        extent=[-300, 0, -300, 0], vmin=vmin, vmax=vmax, origin='lower')
        ax3.set_xlim(-300, 0)
        ax3.set_ylim(-300, 0)
        ax3.set_aspect('equal')
        ax3.set_title('B) Baseline Distance Matrix')
        plt.colorbar(im1, ax=ax3)

        # Response distance matrix
        ax4 = fig.add_subplot(gs[4, 1])
        im2 = ax4.imshow(response_dist, aspect='equal', cmap='viridis',
                        extent=[0, 300, 0, 300], vmin=vmin, vmax=vmax, origin='lower')
        ax4.set_xlim(0, 300)
        ax4.set_ylim(0, 300)
        ax4.set_aspect('equal')
        ax4.set_title('B) Response Distance Matrix')
        plt.colorbar(im2, ax=ax4)

        
        # Add connecting boxes
        for ax in [ax3, ax4]:
            box = plt.Rectangle(
                (ax.get_xlim()[0], ax.get_ylim()[0]),
                ax.get_xlim()[1] - ax.get_xlim()[0],
                ax.get_ylim()[1] - ax.get_ylim()[0],
                fill=False, color='pink', linewidth=1.5
            )
            ax.add_patch(box)
        


        def add_projection_shadow(ax_3d, dist_matrix, is_baseline=True):
            """Create visual connection between matrix and surface."""
            # Get the matrix corner points
            if is_baseline:
                x = np.linspace(-300, 0, dist_matrix.shape[0])
            else:
                x = np.linspace(0, 300, dist_matrix.shape[0])
            X, Y = np.meshgrid(x, x)
            vmin = dist_matrix.min()
            vmax = dist_matrix.max()
            
            # Project on walls and bottom
            ax_3d.contourf(X, Y, dist_matrix, 
                        zdir='z', offset=vmin,
                        levels=np.linspace(vmin, vmax, 20), 
                        cmap='viridis', alpha=0.3)
            
            if is_baseline:
                ax_3d.contourf(X, np.full_like(Y, Y.min()), dist_matrix,
                            zdir='y', offset=Y.min(),
                            levels=np.linspace(vmin, vmax, 20),
                            cmap='viridis', alpha=0.3)
                ax_3d.contourf(np.full_like(X, X.min()), Y, dist_matrix,
                            zdir='x', offset=X.min(),
                            levels=np.linspace(vmin, vmax, 20),
                            cmap='viridis', alpha=0.3)
            else:
                ax_3d.contourf(X, np.full_like(Y, Y.max()), dist_matrix,
                            zdir='y', offset=Y.max(),
                            levels=np.linspace(vmin, vmax, 20),
                            cmap='viridis', alpha=0.3)
                ax_3d.contourf(np.full_like(X, X.max()), Y, dist_matrix,
                            zdir='x', offset=X.max(),
                            levels=np.linspace(vmin, vmax, 20),
                            cmap='viridis', alpha=0.3)

        
        # C: 3D representation
        # Baseline surface
        ax5 = fig.add_subplot(gs[5, 0], projection='3d')
        x = np.linspace(-300, 0, baseline_dist.shape[0])
        X, Y = np.meshgrid(x, x)
        # Main surface plot
        surf1 = ax5.plot_surface(X, Y, baseline_dist, cmap='viridis',
                                vmin=vmin, vmax=vmax, alpha=0.8)
        
        
        # Add threshold plane
        xx, yy = np.meshgrid([-300, 0], [-300, 0])
        zz = np.full_like(X, optimal_threshold)
        ax5.plot_surface(X, Y, zz, alpha=0.2, color='gray')
        
        # Enhanced shadow projections
        ax5.contourf(X, Y, baseline_dist, zdir='z', offset=optimal_threshold,
                    levels=np.linspace(vmin, vmax, 20), cmap='viridis', alpha=0.5)
        
        ax5.set_title('C) Baseline Threshold Surface')
        ax5.view_init(elev=20, azim=-45)
        ax5.set_zlim(vmin, vmax)
        ax5.set_xlim(-300, 0)
        ax5.set_ylim(-300, 0)

        # Response surface
        ax6 = fig.add_subplot(gs[5, 1], projection='3d')
        x = np.linspace(0, 300, response_dist.shape[0])
        X, Y = np.meshgrid(x, x)
        
        # Main surface plot
        surf2 = ax6.plot_surface(X, Y, response_dist, cmap='viridis',
                            vmin=vmin, vmax=vmax, alpha=0.8)
        
        # Add threshold plane
        xx, yy = np.meshgrid([0, 300], [0, 300])
        zz = np.full_like(xx, optimal_threshold)
        ax6.plot_surface(xx, yy, zz, alpha=0.2, color='gray')
        
        ax6.contourf(X, Y, response_dist, zdir='z', offset=optimal_threshold,
                    levels=np.linspace(vmin, vmax, 20), cmap='viridis', alpha=0.5)
        
        ax6.set_title('C) Response Threshold Surface')
        ax6.view_init(elev=20, azim=-45)
        ax6.set_zlim(vmin, vmax)
        ax6.set_xlim(0, 300)
        ax6.set_ylim(0, 300)
        
        fig.canvas.draw()
        
        # Add connecting lines
        #add_connecting_lines(ax3, ax5, fig)
        #add_connecting_lines(ax4, ax6, fig)

        add_projection_shadow(ax5, baseline_dist, is_baseline=True)
        add_projection_shadow(ax6, response_dist, is_baseline=False)

        
        # D: Transition matrices
        # Calculate transition matrices for both thresholds
        baseline_trans_low = self.distance2transition(baseline_dist, lower_threshold)
        response_trans_low = self.distance2transition(response_dist, lower_threshold)
        baseline_trans_opt = self.distance2transition(baseline_dist, optimal_threshold)
        response_trans_opt = self.distance2transition(response_dist, optimal_threshold)

        # Get maximum size to pad all matrices to same size
        max_size = max(baseline_trans_low.shape[0], 
                    response_trans_low.shape[0],
                    baseline_trans_opt.shape[0], 
                    response_trans_opt.shape[0])

        # Function to pad matrix to target size
        def pad_matrix(matrix, target_size):
            pad_width = target_size - matrix.shape[0]
            if pad_width > 0:
                return np.pad(matrix, ((0, pad_width), (0, pad_width)), mode='constant')
            return matrix

        # Pad all matrices to same size
        baseline_trans_low = pad_matrix(baseline_trans_low, max_size)
        response_trans_low = pad_matrix(response_trans_low, max_size)
        baseline_trans_opt = pad_matrix(baseline_trans_opt, max_size)
        response_trans_opt = pad_matrix(response_trans_opt, max_size)

        # Calculate sizes for the grid
        n = max_size
        full_size = n * 2  # Total size for 2x2 grid

        # Create empty array for full grid
        full_trans = np.zeros((full_size, full_size))

        # Fill in the quadrants
        full_trans[:n, :n] = baseline_trans_low
        full_trans[:n, n:] = baseline_trans_opt
        full_trans[n:, :n] = response_trans_low
        full_trans[n:, n:] = response_trans_opt
        
        # Create a new subplot for the consolidated transition matrices
        ax_trans = fig.add_subplot(gs[6, :])  # Use one row instead of two
        
        # Plot the consolidated matrix
        im = ax_trans.imshow(full_trans, cmap='binary')
        
        # Add grid lines to separate quadrants
        ax_trans.axhline(y=n-0.5, color='r', linestyle='-', linewidth=0.5)
        ax_trans.axvline(x=n-0.5, color='r', linestyle='-', linewidth=0.5)
        
        # Add labels
        ax_trans.text(-n/4, n/2, f'ε\' = {lower_threshold:.3f}', 
                    rotation=90, ha='center', va='center')
        ax_trans.text(-n/4, n*1.5, f'ε* = {optimal_threshold:.3f}', 
                    rotation=90, ha='center', va='center')
        ax_trans.text(n/2, -n/4, 'Baseline', ha='center', va='center')
        ax_trans.text(n*1.5, -n/4, 'Response', ha='center', va='center')
        ax_trans.set_title('D) Transition Matrices')
        
        # E: NST analysis
        ax11 = fig.add_subplot(gs[7, :])
        ax11.plot(details['thresholds'][0], details['nst_resp_values'][0],
                'b-', label='NST Response (NSTres)')
        ax11.plot(details['thresholds'][0], details['nst_base_values'][0],
                'k-', label='NST Baseline (NSTbase)')
        ax11.plot(details['thresholds'][0], details['diff_values'][0],
                'r-', label='ΔNST = NSTres - k × NSTbase')
        
        # Add both threshold lines
        ax11.axvline(x=lower_threshold, color='gray', linestyle='--', label='ε\'')
        ax11.axvline(x=optimal_threshold, color='r', linestyle='--', label='ε*')
        
        ax11.set_xlabel('Threshold (ε)')
        ax11.set_ylabel('Number of State Transitions')
        ax11.legend()
        ax11.set_title('E) State Transitions')
        
        # Add PCIst equation
        ax11.text(0.95, 0.95, 'PCIst = ∑ΔNSTn', transform=ax11.transAxes,
                ha='right', va='top', fontsize=10)
        
        plt.subplots_adjust(hspace=0.4)
        plt.tight_layout()
        return fig
            




from pathlib import Path

import numpy as np
import pytest
from mne import read_epochs_eeglab
from mne.io import read_raw
from scipy.io import loadmat

from tmseegpy.mne_icalabel.datasets import icalabel
from tmseegpy.mne_icalabel.iclabel._utils import _gdatav4, _mne_to_eeglab_locs, _next_power_of_2

dataset_path = icalabel.data_path() / "iclabel"

# Raw/Epochs files with ICA decomposition
raw_eeglab_path = dataset_path / "datasets/sample-raw.set"
epo_eeglab_path = dataset_path / "datasets/sample-epo.set"

# Electrode locations
loc_raw_path = dataset_path / "utils/loc-raw.mat"
loc_epo_path = dataset_path / "utils/loc-raw.mat"

# Grid data interpolation
gdatav4_raw_path = dataset_path / "utils/gdatav4-raw.mat"
gdatav4_epo_path = dataset_path / "utils/gdatav4-epo.mat"


# General readers
reader = {"raw": read_raw, "epo": read_epochs_eeglab}
kwargs = {"raw": dict(preload=True), "epo": dict()}


@pytest.mark.filterwarnings("ignore:Estimated head radius.*:RuntimeWarning")
@pytest.mark.parametrize(
    ("file", "eeglab_result_file"),
    [(raw_eeglab_path, loc_raw_path), (epo_eeglab_path, loc_epo_path)],
)
def test_loc(file, eeglab_result_file):
    """Test conversion of MNE montage to EEGLAB loc.

    This test works because MNE does the conversion from EEGLAB to MNE montage
    when loading the datasets.
    """
    type_ = str(Path(file).stem)[-3:]
    inst = reader[type_](file, **kwargs[type_])
    rd, th = _mne_to_eeglab_locs(inst, picks=inst.ch_names)
    eeglab_loc = loadmat(eeglab_result_file)["loc"][0, 0]
    eeglab_rd = eeglab_loc["rd"]
    eeglab_th = eeglab_loc["th"]
    assert np.allclose(rd, eeglab_rd, atol=1e-8)
    assert np.allclose(th, eeglab_th, atol=1e-8)


@pytest.mark.parametrize("file", [gdatav4_raw_path, gdatav4_epo_path])
def test_gdatav4(file):
    """Test grid data interpolation."""
    # load inputs from MATLAB
    eeglab_gdata = loadmat(file)["gdatav4"][0, 0]
    eeglab_inty = eeglab_gdata["inty"]
    eeglab_intx = eeglab_gdata["intx"]
    eeglab_intValues = eeglab_gdata["intValues"]
    eeglab_yi = eeglab_gdata["yi"]
    eeglab_xi = eeglab_gdata["xi"]

    # create mesh
    xq, yq = np.meshgrid(eeglab_xi, eeglab_yi)

    # compute output in Python
    Xi, Yi, Zi = _gdatav4(eeglab_intx, eeglab_inty, eeglab_intValues, xq, yq)

    # load outputs from MATLAB
    eeglab_Xi = eeglab_gdata["Xi"]
    eeglab_Yi = eeglab_gdata["Yi"]
    eeglab_Zi = eeglab_gdata["Zi"]

    # compare
    assert np.allclose(Xi, eeglab_Xi, atol=1e-8)
    assert np.allclose(Yi, eeglab_Yi, atol=1e-8)
    # Zi has to be transposed in Python
    assert np.allclose(Zi.T, eeglab_Zi, atol=1e-8)


def test_next_power_of_2():
    """Test that next_power_of_2 works as intended."""
    x = [0, 10, 200, 400]
    expected = [1, 16, 256, 512]
    for k, exp in zip(x, expected):
        val = _next_power_of_2(k)
        assert exp == val



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#  This file (__init__.py) is part of neurone_loader                           -
#  (https://www.github.com/heilerich/neurone_loader)                           -
#  Copyright © 2019 Felix Heilmeyer.                                           -
#                                                                              -
#  This code is released under the MIT License                                 -
#  https://opensource.org/licenses/mit-license.php                             -
#  Please see the file LICENSE for details.                                    -
# ------------------------------------------------------------------------------
"""
Provides classes to load, represent and export data recorded with the Bittium NeurOne device.
Use :py:class:`Recording`, :py:class:`Session` or :py:class:`Phase` to represent the respective part of the data.
"""

from .loader import Recording, Session, Phase

__version__ = '2.0.1'



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#  This file (neurone.py) is part of neurone_loader                            -
#  (https://www.github.com/heilerich/neurone_loader)                           -
#  Copyright © 2019 Felix Heilmeyer.                                           -
#                                                                              -
#  This code is released under the MIT License                                 -
#  https://opensource.org/licenses/mit-license.php                             -
#  Please see the file LICENSE for details.                                    -
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# This file contains code originally from export2hdf5                          -
# (https://github.com/bwrc/export2hdf5)                                        -
# Created by Andreas Henelius <andreas.henelius@ttl.fi>,                       -
# Finnish Institute of Occupational Health                                     -
# ------------------------------------------------------------------------------

"""
Contains functions for reading data recorded with a
Bittium NeurOne device. This module currently supports
reading of data and events.
"""

import numpy as np
import xml.etree.ElementTree
import re

from os import path
from construct import Struct, Int32sl, Int64ul

from datetime import datetime

from collections import namedtuple


def read_neurone_protocol(fpath):
    """
    Read the measurement protocol from an XML file.

    Arguments:
       - fpath: the path to the directory holding the
                 NeurOne measurement (i.e., the
                 directory Protocol.xml and Session.xml
                 files.

    Returns:
       - a dictionary containing (i) the names of the channels
         in the recording and (ii) meta information
         (recording start/stop times, sampling rate).

    {"meta" : <dict with metadata>,
    "channels" : <array with channel names>}
    """

    # Define filename
    fname_protocol = path.join(fpath, "Protocol.xml")
    fname_session = path.join(fpath, "Session.xml")

    # --------------------------------------------------
    # Read the protocol data
    # --------------------------------------------------
    # Define the XML namespace as a shorthand
    ns = {'xmlns': 'http://www.megaemg.com/DataSetGeneralProtocol.xsd'}

    # Get channel names and organise them according to their
    # physical order (InputNumber), which is the order
    # in which the channels are being sampled.
    doc_root = xml.etree.ElementTree.parse(fname_protocol).getroot()
    channels = doc_root.findall("xmlns:TableInput", namespaces=ns)
    channel_names = [(0, 0)] * len(channels)

    for i, ch in enumerate(channels):
        channel_names[i] = (int(ch.findall("xmlns:PhysicalInputNumber", namespaces=ns)[0].text),
                            ch.findall("xmlns:Name", namespaces=ns)[0].text)
    channel_names = [i for _, i in sorted(channel_names)]

    # Get the sampling rate
    sampling_rate = int(doc_root.findall("xmlns:TableProtocol", namespaces=ns)[0]
                        .findall("xmlns:ActualSamplingFrequency", namespaces=ns)[0].text)

    # --------------------------------------------------
    # Read the session data
    # --------------------------------------------------
    # Define the XML namespace as a shorthand
    ns2 = {'xmlns': 'http://www.megaemg.com/DataSetGeneralSession.xsd'}

    # Get channel names and organise them according to their
    # physical order (InputNumber), which is the order
    # in which the channels are being sampled.
    doc_root = xml.etree.ElementTree.parse(fname_session).getroot()
    session = doc_root.findall("xmlns:TableSession", namespaces=ns2)
    time_start = session[0].findall("xmlns:StartDateTime", namespaces=ns2)[0].text
    time_stop = session[0].findall("xmlns:StopDateTime", namespaces=ns2)[0].text
    phases = [{'number': phase.findall("xmlns:Folder", namespaces=ns2)[0].text.split("\\")[-1],
               'time_start': phase.findall("xmlns:StartDateTime", namespaces=ns2)[0].text,
               'time_stop': phase.findall("xmlns:StopDateTime", namespaces=ns2)[0].text}
              for phase in doc_root.findall("xmlns:TableSessionPhase", namespaces=ns2)]

    subject_info = doc_root.find('xmlns:TablePerson', namespaces=ns2)
    subject_id = subject_info.find('xmlns:PersonID', namespaces=ns2).text
    subject_first_name = subject_info.find('xmlns:FirstName', namespaces=ns2).text
    subject_last_name = subject_info.find('xmlns:LastName', namespaces=ns2).text
    subject_dob = subject_info.find('xmlns:DateOfBirth', namespaces=ns2).text

    # --------------------------------------------------
    # Package the information
    # --------------------------------------------------
    meta = {}

    meta["time_start"] = _convert_time(time_start)
    meta["time_stop"] = _convert_time(time_stop)
    meta["sampling_rate"] = sampling_rate

    meta["subject"] = dict(
        id=subject_id,
        first_name=subject_first_name,
        last_name=subject_last_name,
        date_of_birth=_convert_time(subject_dob)
    )
    
    for phase in phases:
        phase['time_start'] = _convert_time(phase['time_start'])
        phase['time_stop'] = _convert_time(phase['time_stop'])

    return {'channels': channel_names, 'meta': meta, 'phases': phases}

from datetime import datetime
import re

def _convert_time(inp_str):
    """
    Converts a datetime string to a datetime object.
    Handles both naive and timezone-aware datetime strings.
    """
    try:
        # Handle 'Z' timezone
        if inp_str.endswith('Z'):
            # Replace 'Z' with '+00:00' for UTC
            inp_str = inp_str[:-1] + '+00:00'
        
        # Use regex to parse the datetime components
        iso_format_regex = r'^(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2})(\.\d+)?([+-]\d{2}:\d{2})?$'
        match = re.match(iso_format_regex, inp_str)
        if match:
            main_time, fractional, timezone = match.groups()
            if fractional:
                # Remove the dot
                fractional = fractional[1:]
                # Limit to 6 digits by truncating or rounding
                fractional = fractional[:6].ljust(6, '0')
                microsecond = int(fractional)
            else:
                microsecond = 0
            # Reconstruct datetime string without fractional seconds
            dt_str = main_time
            if timezone:
                dt_str += timezone
            dt = datetime.fromisoformat(dt_str)
            dt = dt.replace(microsecond=microsecond)
            return dt
        else:
            raise ValueError("Input string does not match expected format.")
    except Exception as e:
        raise ValueError(f"Error parsing time string '{inp_str}': {e}")


def read_neurone_data(fpath, session_phase=1, protocol=None):
    """
    Read the NeurOne signal data from a binary file.

    Arguments:
       - fpath: the path to the directory holding the
                 NeurOne measurement (i.e., the
                 directory Protocol.xml and Session.xml
                 files.

       - session_phase:
                 The phase of the measurement. Currently
                 only reading of the first phase (1) is
                 supported.

       - protocol:
                  The dictionary obtained using the function
                  read_neurone_protocol. This argument is optional
                  and if not given, the protocol is automatically read.
                    
    Returns:
       - A numpy ndarray with the data, where each columns stores
         the data for one channel.
    """

    fname = path.join(fpath, str(session_phase), '1.bin')

    # Read the protocol unless provided
    if protocol is None:
        protocol = read_neurone_protocol(fpath)
    
    # Determine number of samples to read
    n_samples, n_channels = read_neurone_data_info(fpath, session_phase, protocol)

    # Read the data and store the data
    # in an ndarray
    data = np.fromfile(fname, dtype='<i4')
    data.shape = (n_samples, n_channels)

    return data


def read_neurone_data_info(fpath, session_phase=1, protocol=None):
    """
    Read the sample and channel count from a NeurOne signal binary file.

    Arguments:
       - fpath: the path to the directory holding the
                 NeurOne measurement (i.e., the
                 directory Protocol.xml and Session.xml
                 files.

       - session_phase:
                 The phase of the measurement. Currently
                 only reading of the first phase (1) is
                 supported.

       - protocol:
                  The dictionary obtained using the function
                  read_neurone_protocol. This argument is optional
                  and if not given, the protocol is automatically read.
                    
    Returns:
       Returns:
       - a named tuple containing (i) the number of channels
         and (ii) the number of samples in the recording.

        ( n_samples, n_channels )
    """

    fname = path.join(fpath, str(session_phase), '1.bin')

    # Read the protocol unless provided
    if protocol is None:
        protocol = read_neurone_protocol(fpath)
    
    # Determine number of samples and channels
    f_info = path.getsize(fname)
    n_channels = len(protocol['channels'])
    n_samples = int(f_info / 4 / n_channels)
    
    DataInfo = namedtuple('DataInfo', ['n_samples', 'n_channels'])
    return DataInfo(n_samples, n_channels)


def get_n1_event_format():
    """
    Define the format for the events in a neurone recording.
    
    Arguments: None.

    Returns:
       - A Struct (from the construct library) describing the
         event format.
    """

    # Define the data format of the events
    # noinspection PyUnresolvedReferences
    return Struct(
        "Revision" / Int32sl,
        "RFU1" / Int32sl,
        "Type" / Int32sl,
        "SourcePort" / Int32sl,
        "ChannelNumber" / Int32sl,
        "Code" / Int32sl,
        "StartSampleIndex" / Int64ul,
        "StopSampleIndex" / Int64ul,
        "DescriptionLength" / Int64ul,
        "DescriptionOffset" / Int64ul,
        "DataLength" / Int64ul,
        "DataOffset" / Int64ul,
        "RFU2" / Int32sl,
        "RFU3" / Int32sl,
        "RFU4" / Int32sl,
        "RFU5" / Int32sl
    )


def read_neurone_events(fpath, session_phase=1, sampling_rate=None):
    """
    Read the NeurOne events from a binary file.

    Arguments:
       - fpath: the path to the directory holding the
                 NeurOne measurement (i.e., the
                 directory Protocol.xml and Session.xml
                 files.

       - sampling_rate:
                 The sampling rate of the recording.
                 This argument is optional and if not given,
                 the protocol is automatically read.

       - session_phase:
                 The phase of the measurement. Currently
                 only reading of the first phase (1) is
                 supported.

    Returns:
       - A dict containing the events and the data type for the events.
    {"events" : <numpy structured array with the events>,
    "events_dtype" : <array with the numpy dtype for the events>}
    """

    fname = path.join(fpath, str(session_phase), "events.bin")

    # Get the sampling rate unless provided
    if sampling_rate is None:
        protocol = read_neurone_protocol(fpath)
        sampling_rate = protocol['meta']['sampling_rate']
    
    # Determine number of events
    f_info = path.getsize(fname)
    n_events = int(f_info / 88)
    events = [{}] * n_events

    # Read events in chunks of 88 bytes and unpack
    # also add start / stop time for each event
    # and remove 'reserved for future use' (RFU) fields
    event_format = get_n1_event_format()
    with open(fname, mode='rb') as file:
        for i in range(n_events):
            events[i] = event_format.parse(file.read(88))
            events[i]['StartTime'] = events[i]['StartSampleIndex'] / sampling_rate
            events[i]['StopTime'] = events[i]['StopSampleIndex'] / sampling_rate
            for j in range(5):
                del events[i]['RFU' + str(j+1)]
            del events[i]['_io']

    # Create a numpy structured array from the events
    events_dtype = np.dtype([("Revision", np.int32),
                             ("Type", np.int32),
                             ("SourcePort", np.int32),
                             ("ChannelNumber", np.int32),
                             ("Code", np.int32),
                             ("StartSampleIndex", np.int64),
                             ("StopSampleIndex", np.int64),
                             ("DescriptionLength", np.int64),
                             ("DescriptionOffset", np.int64),
                             ("DataLength", np.int64),
                             ("DataOffset", np.int64),
                             ("StartTime", np.int64),
                             ("StopTime", np.int64)])

    # convert array of event dicts to an array of tuples
    if len(events) == 0:
        return {'events': np.array([], dtype=events_dtype), 'dtype': events_dtype}
    key_list = [k for k, v in events[0].items()]
    tmp = [tuple([e[k] for k in key_list]) for e in events]
    events = np.array(tmp, dtype=events_dtype)

    return {'events': events, 'dtype': events_dtype}



from __future__ import annotations  # c.f. PEP 563, PEP 649

import shutil
from functools import partial
from typing import TYPE_CHECKING

import pooch
from mne.datasets import fetch_dataset
from mne.datasets.utils import has_dataset
from mne.utils import verbose

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Optional

has_icalabel_testing_data = partial(has_dataset, name="icalabel-testing")


@verbose
def data_path(
    path: Optional[str] = None,
    force_update: bool = False,
    update_path: bool = True,
    download: bool = True,
    verbose=None,
) -> Path:
    """ICA label testing data generated in conjunction with EEGLab.

    Parameters
    ----------
    path : None | str
        Location of where to look for the dataset.
        If None, the environment variable or config parameter is used.
        If it doesn’t exist, the “~/mne_data” directory is used.
        If the dataset is not found under the given path,
        the data will be automatically downloaded to the specified folder.
    force_update : bool
        Force update of the dataset even if a local copy exists.
    update_path : bool
        If True, set the MNE_DATASETS_FNIRSMOTORGROUP_PATH in
        mne-python config to the given path.
    download : bool
        If False and the dataset has not been downloaded yet,
        it will not be downloaded and the path will be returned
        as ``‘’`` (empty string). This is mostly used for debugging purposes
        and can be safely ignored by most users.
    %(verbose)s

    Returns
    -------
    path : Path
        Path to dataset directory.
    """
    dataset_params = dict(
        archive_name="MNE-testing-icalabel-data.zip",
        hash="md5:ea1b188e671fe35a65c2c3cb6c42ccb3",
        url="https://github.com/adam2392/mne-testing-icalabel-data/archive/main.zip",
        folder_name="MNE-testing-icalabel-data",
        dataset_name="icalabel-testing",
        config_key="MNE_ICALABEL_DATASET_TESTING_PATH",
    )
    folder_name = dataset_params["folder_name"]

    dpath = fetch_dataset(
        dataset_params,
        path=path,
        force_update=force_update,
        update_path=update_path,
        download=download,
        processor=pooch.Unzip(extract_dir=f"./{folder_name}"),
    )

    # do some wrangling to deal with nested directories
    bad_name = dpath / "mne-testing-icalabel-data-main"
    if bad_name.is_dir():
        shutil.move(bad_name, dpath.with_suffix(".true"))
        shutil.rmtree(dpath)
        shutil.move(dpath.with_suffix(".true"), dpath)

    return dpath



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#  This file (mne_export.py) is part of neurone_loader                         -
#  (https://www.github.com/heilerich/neurone_loader)                           -
#  Copyright © 2019 Felix Heilmeyer.                                           -
#                                                                              -
#  This code is released under the MIT License                                 -
#  https://opensource.org/licenses/mit-license.php                             -
#  Please see the file LICENSE for details.                                    -
# ------------------------------------------------------------------------------
"""
Provides the metaclass `MneExportable` that allows subclasses implementing all the metaclass's properties
to be converted to a `mne.io.RawArray`.
"""

import abc
from datetime import datetime, timezone, timedelta
from copy import deepcopy

import numpy as np

from .util import logger


class UnknownChannelException(Exception):
    """

    Raised if data contains a channel name that is neither in a list of well-known channels nor in an (optional) list
    of user supplied channel name to channel type mappings.

    """
    pass


class MneExportable(abc.ABC):
    """

    A metaclass that provides a function allowing objects that expose data, events, channels and sampling_rate
    properties to be converted to an mne.io.RawArray.

    """

    def _import_mne(self):
        try:
            # noinspection PyPackageRequirements
            import mne
            self._mne = mne
            return True
        except ImportError:
            logger.error("To convert data to an MNE object you must install MNE "
                         "and all its dependencies.")
            return False


    def to_mne(self, substitute_zero_events_with=None, copy=False, channel_type_mappings=None):
        """
        Convert loaded data to a mne.io.RawArray

        :param substitute_zero_events_with: None. events with code = 0 are not supported by MNE, if this parameter is
                                            set, the event code 0 will be substituted with this parameter
        :param copy: False. If False (default), the original data will be removed from memory to save space while
                     creating the mne.io.RawArray. If the data is needed again it must be reloaded from disk
        :param channel_type_mappings: Optional. You can provide a dictionary of channel name to type mappings. If the
                                      data contains any channel not in the list of well-known channel names and not in
                                      this mapping the conversion will raise UnknownChannelException. You can choose
                                      to map any unknown channel to one specific type e.g. {'#unknown': 'eeg'}. For a
                                      list of available types see the documentation of :func:`mne.pick_types`. This
                                      setting takes precedence over the built-in list of common channel names.
        :type substitute_zero_events_with: None or int
        :type copy: bool
        :type channel_type_mappings: None or dict
        :return: the converted data
        :rtype: mne.io.RawArray
        :raises ImportError: if the mne package is not installed
        :raises UnknownChannelException: if a unknown channel name is encountered (see channel_type_mappings parameter)
        """
        from datetime import datetime, timezone, timedelta, date
        from copy import deepcopy
        import numpy as np

        if not hasattr(self, '_mne'):
                if not self._import_mne():
                    raise ImportError
        mne = self._mne

        allowed_channel_types = ['meg', 'eeg', 'stim', 'eog', 'ecg', 'emg', 'misc', 'resp', 'chpi', 'exci', 'ias',
                                 'syst', 'seeg', 'dipole', 'gof', 'bio', 'ecog', 'fnirs']
        if channel_type_mappings is not None:
            for value in channel_type_mappings.values():
                assert value in allowed_channel_types, '{type} is not a recognized mne channel type'.format(type=value)

        events = self.events

        def _channel_type(name):
            mappings = [
                (['emg'], 'emg'),
                (['eog'], 'eog'),
                (['ecg'], 'ecg'),
                (['Microphone', 'GSR'], 'bio'),
                (_default_eeg_channel_names, 'eeg')
            ]

            def _get_common_type(ch_name):
                for starts, ch_type in mappings:
                    for start in starts:
                        if ch_name.lower().startswith(start.lower()):
                            return ch_type
                return None

            if channel_type_mappings is not None:
                if name in channel_type_mappings:
                    assert channel_type_mappings[name] in allowed_channel_types, '{type} is not a recognized mne' \
                                                                                 ' channel type'.format(type=value)
                    return channel_type_mappings[name]
                elif _get_common_type(name) is not None:
                    return _get_common_type(name)
                elif '#unknown' in channel_type_mappings:
                    unknown_channel_type = channel_type_mappings['#unknown']
                    logger.warning('Could not properly classify {channel}. It will be classified as {default} channel '
                                   'because {default} was set as default for unknown channels.'
                                   .format(channel=name, default=unknown_channel_type))
                    return unknown_channel_type
            else:
                if _get_common_type(name) is not None:
                    return _get_common_type(name)

            logger.error('Encountered channel ({channel}) which is not in the list of well-known channels'
                         'and no user defined mapping was supplied.'.format(channel=name))
            raise UnknownChannelException

        channel_types = list(map(_channel_type, self.channels))
        assert len(channel_types) == len(self.channels)

        data_length = self.data.T.shape[1]
        data_dtype = self.data.dtype

        ssc = [(start, stop, code) for (start, stop, code)
               in events[['StartSampleIndex', 'StopSampleIndex', 'Code']].values]

        if substitute_zero_events_with is not None:
            event_codes = np.unique(events['Code'].values)
            assert type(substitute_zero_events_with) is int, 'substitute_zero_events_with must be int or None'
            assert substitute_zero_events_with not in event_codes, \
                """the original data can't contain event with code
                substitute_zero_events_with ({})""".format(substitute_zero_events_with)

        stim_channel_names = ['STI 014']
        stim_channels = [np.zeros(data_length, dtype=data_dtype)]

        def _add_stim_channel():
            index = len(stim_channel_names)
            stim_channel_names.append('STI {:03g}'.format(index + 14))
            stim_channels.append(np.zeros(data_length, dtype=data_dtype))

        for start, stop, code in ssc:
            channel_index = 0
            if substitute_zero_events_with is not None:
                if code == 0:
                    code = substitute_zero_events_with
            else:
                assert code != 0, "events with event code 0 are not supported by MNE, use the " \
                                  "substitute_zero_events_with parameter of this method to substitute with an " \
                                  "alternative code"

            corrected_start = start - 1 if start != 0 else 0  # sample before event start needs to be 0
            while not (stim_channels[channel_index][corrected_start:stop + 1] == 0).all():
                channel_index += 1
                if channel_index >= len(stim_channels):
                    _add_stim_channel()

            if channel_index != 0:
                logger.warning('Event (code {current_code}) has a concurrent event between samples ({start}) and '
                               '({stop}). An additional stim channel will be used/created: ({channel}).'
                               .format(current_code=code, start=start, stop=stop,
                                       channel=stim_channel_names[channel_index]))

            stim_channels[channel_index][start:stop + 1] += code  # event must be at least one sample long

        stim_info = mne.create_info(ch_names=stim_channel_names,
                                    sfreq=self.sampling_rate,
                                    ch_types='stim')

        stim_cnt = mne.io.RawArray(stim_channels, stim_info, verbose='WARNING')

        channel_names = deepcopy(self.channels)
        channel_names.extend(stim_channel_names)
        all_channel_types = deepcopy(channel_types)
        all_channel_types.extend(['stim' for _ in stim_channel_names])

        info = mne.create_info(ch_names=self.channels,
                               sfreq=self.sampling_rate,
                               ch_types=channel_types)

        # data is µV samples x channels, mne needs V channels x samples
        data = self.data.T / (1000 * 1000)
        cnt = mne.io.RawArray(data, info)
        if not copy:
            self.clear_data()

        cnt = cnt.add_channels([stim_cnt])

        if self.subject_info is not None:
            mne_sub_info = dict()
            mne_sub_info['his_id'] = self.subject_info['id']
            mne_sub_info['last_name'] = self.subject_info['last_name']
            mne_sub_info['first_name'] = self.subject_info['first_name']
            
            # Modified birthday handling
            dob = self.subject_info['date_of_birth']
            if dob is not None:
                if isinstance(dob, (tuple, list)):
                    # Convert tuple/list to date object
                    try:
                        dob = date(year=dob[0], month=dob[1], day=dob[2])
                    except (IndexError, TypeError):
                        dob = None
                elif isinstance(dob, str):
                    # Try to parse string date
                    try:
                        dob = datetime.strptime(dob, '%Y-%m-%d').date()
                    except ValueError:
                        dob = None
                elif isinstance(dob, datetime):
                    dob = dob.date()
                elif not isinstance(dob, date):
                    dob = None
                    
                mne_sub_info['birthday'] = dob

            cnt.info['subject_info'] = mne_sub_info

        if self.time_start is not None:
            tstart = self.time_start
            if tstart.tzinfo is not None:
                # Convert to UTC timezone
                tstart = tstart.astimezone(timezone.utc)
            else:
                # If no timezone info, assume UTC
                tstart = tstart.replace(tzinfo=timezone.utc)
            # Set meas_date with the timezone-aware datetime
            cnt.set_meas_date(tstart)

        return cnt

    @property
    @abc.abstractmethod
    def subject_info(self):
        """
        Abstract Property

        :return: should contain subject_info dict with keys `id`, `first_name`, `last_name`, `date_of_birth`
        :rtype: dict
        """

    @property
    @abc.abstractmethod
    def time_start(self):
        """
        Abstract Property

        :return: should contain start time of recording
        :rtype: datetime
        """

    @property
    @abc.abstractmethod
    def data(self):
        """
        Abstract Property

        :return: should contain data in (n_samples, n_channels) shape
        :rtype: numpy.ndarray
        """

    @abc.abstractmethod
    def clear_data(self):
        """
        Abstract Method

        Should delete loaded data from memory
        """

    @property
    @abc.abstractmethod
    def events(self):
        """
        Abstract Property

        :return: should contain the events as a DataFrame, required fields are `StartSampleIndex`, `StopSampleIndex`
                 and `Code`. Additional fields are ignored.
        :rtype: pandas.DataFrame
        """

    @property
    @abc.abstractmethod
    def channels(self):
        """
        Abstract Property

        :return: should contain the names of channels, matching the sequence used in the data property
        :rtype: list[str]
        """

    @property
    @abc.abstractmethod
    def sampling_rate(self):
        """
        Abstract Property

        :return: should contain the used sampling rate
        :rtype: int
        """


_default_eeg_channel_names = ['Fp1', 'Fpz', 'Fp2', 'F7', 'F3', 'Fz', 'F4', 'F8', 'FC5', 'FC1', 'FC2', 'FC6', 'M1', 'T7',
                              'C3', 'Cz', 'C4', 'T8', 'M2', 'CP5', 'CP1', 'CP2', 'CP6', 'P7', 'P3', 'Pz', 'P4', 'P8',
                              'POz', 'O1', 'Oz', 'O2', 'AF7', 'AF3', 'AF4', 'AF8', 'F5', 'F1', 'F2', 'F6', 'FC3', 'FCz',
                              'FC4', 'C5', 'C1', 'C2', 'C6', 'CP3', 'CPz', 'CP4', 'P5', 'P1', 'P2', 'P6', 'PO5', 'PO3',
                              'PO4', 'PO6', 'FT7', 'FT8', 'TP7', 'TP8', 'TP9', 'PO7', 'PO8', 'FT9', 'FT10', 'TPP9h', 'TPP10h',
                              'PO9', 'PO10', 'P9', 'P10', 'AFF1', 'AFz', 'AFF2', 'FFC5h', 'FFC3h', 'FFC4h', 'FFC6h',
                              'FCC5h', 'FCC3h', 'FCC4h', 'FCC6h', 'CCP5h', 'CCP3h', 'CCP4h', 'CCP6h', 'CPP5h', 'CPP3h',
                              'CPP4h', 'CPP6h', 'PPO1', 'PPO2', 'I1', 'Iz', 'I2', 'AFp3h', 'AFp4h', 'AFF5h', 'AFF6h',
                              'FFT7h', 'FFC1h', 'FFC2h', 'FFT8h', 'FTT9h', 'FTT7h', 'FCC1h', 'FCC2h', 'FTT8h', 'FTT10h',
                              'TTP7h', 'CCP1h', 'CCP2h', 'TTP8h', 'TPP7h', 'CPP1h', 'CPP2h', 'TPP8h', 'PPO9h', 'PPO5h',
                              'PPO6h', 'PPO10h', 'POO9h', 'POO3h', 'POO4h', 'POO10h', 'OI1h', 'OI2h', 'TP9', 'TP10', 'FCz', 'EMG1']
